from datetime import date

import pandas as pd
from django.utils import timezone
from faker import Faker
from pandas.tseries.offsets import BusinessMonthEnd
from wbfdm.factories import (
    CashFactory,
    ClassificationFactory,
    ClassificationGroupFactory,
    EquityFactory,
    InstrumentFactory,
)
from wbportfolio.factories import (
    AssetPositionFactory,
    InstrumentPriceFactory,
    PortfolioFactory,
    PortfolioSynchronizationFactory,
    PriceComputationFactory,
    ProductFactory,
)
from wbportfolio.models import (
    InstrumentPortfolioThroughModel,
    PortfolioPortfolioThroughModel,
)

fake = Faker()


def setup_product(today):
    product = ProductFactory.create()
    product.related_instruments.add(InstrumentFactory.create())
    primary_group = ClassificationGroupFactory.create(is_primary=True, max_depth=3)
    classification_parent_1 = ClassificationFactory.create(group=primary_group)
    classification_parent_2 = ClassificationFactory.create(group=primary_group)

    i1 = EquityFactory.create(classifications=[classification_parent_1.children.first().children.first()])
    i2 = EquityFactory.create(classifications=[classification_parent_2.children.first().children.first()])
    i3 = CashFactory.create()
    for _d in pd.date_range(
        today - BusinessMonthEnd(1), today + BusinessMonthEnd(0), freq="B"
    ):  # Build a complete factsheet month
        _d = _d.date()
        InstrumentPriceFactory.create(instrument=product, date=_d)
        pos = [
            AssetPositionFactory.create(
                portfolio=product.portfolio,
                date=_d,
                underlying_instrument=i1,
                underlying_instrument_price=InstrumentPriceFactory.create(instrument=i1, date=_d, calculated=False),
            ),
            AssetPositionFactory.create(
                portfolio=product.portfolio,
                date=_d,
                underlying_instrument=i2,
                underlying_instrument_price=InstrumentPriceFactory.create(instrument=i2, date=_d, calculated=False),
            ),
            AssetPositionFactory.create(
                portfolio=product.portfolio,
                date=_d,
                underlying_instrument=i3,
                underlying_instrument_price=InstrumentPriceFactory.create(instrument=i3, date=_d, calculated=False),
            ),
        ]
        for p in pos:
            InstrumentPriceFactory.create(instrument=p.underlying_instrument, date=_d)
    return product


class ValidTodayProductSetupMixin:
    def setup_method(self, method=None):
        product = setup_product(timezone.now().date())
        self.product = product


class ValidTodayMultiThematicProductSetupMixin:
    def setup_method(self, method=None):
        today = date.today()
        main_product = ProductFactory.create(price_computation=PriceComputationFactory.create())
        main_product.related_instruments.add(InstrumentFactory.create())

        # Create two valid products that will span our portfolio
        product1 = setup_product(today)
        product2 = setup_product(today)

        # Create a computed portfolio (model to allow automatic sync), linked throught he primary portfolio with a Thematic relationship
        computed_portfolio = PortfolioFactory.create(
            portfolio_synchronization=PortfolioSynchronizationFactory.create(),
        )
        PortfolioPortfolioThroughModel.objects.create(
            portfolio=computed_portfolio, dependency_portfolio=main_product.primary_portfolio
        )

        InstrumentPortfolioThroughModel.objects.update_or_create(
            instrument=main_product, defaults={"portfolio": computed_portfolio}
        )

        # Loop through the date range and create an equally weighted portfolio of products and synchronize the computed portfolio and compute the estimate NAV
        for _d in pd.date_range(today - BusinessMonthEnd(1), today + BusinessMonthEnd(0), freq="B"):
            _d = _d.date()
            AssetPositionFactory.create(
                portfolio=main_product.primary_portfolio,
                underlying_instrument=product1,
                date=_d,
                weighting=0.5,
                initial_price=product1.prices.get(date=_d).net_value,
            )
            AssetPositionFactory.create(
                portfolio=main_product.primary_portfolio,
                underlying_instrument=product2,
                date=_d,
                weighting=0.5,
                initial_price=product2.prices.get(date=_d).net_value,
            )
            computed_portfolio.portfolio_synchronization.synchronize(
                computed_portfolio, _d, override_execution_datetime_validity=True
            )
            main_product.price_computation.compute(main_product, _d, override_execution_datetime_validity=True)
        main_product.prices.update(calculated=False)
        self.product = main_product
        self.theme_portfolio = main_product.primary_portfolio



from wbcore import viewsets
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbportfolio.models import PortfolioSynchronization, PriceComputation
from wbportfolio.serializers import (
    PortfolioSynchronizationRepresentationSerializer,
    PriceComputationRepresentationSerializer,
)


class PortfolioSynchronizationRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:portfoliosynchronization"

    ordering_fields = ordering = ("name",)
    search_fields = ("name",)
    queryset = PortfolioSynchronization.objects.all()
    serializer_class = PortfolioSynchronizationRepresentationSerializer


class PriceComputationRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:portfoliosynchronization"

    ordering_fields = ordering = ("name",)
    search_fields = ("name",)
    queryset = PriceComputation.objects.all()
    serializer_class = PriceComputationRepresentationSerializer



from datetime import datetime

from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.currency.models import Currency
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.date import get_date_interval_from_request
from wbportfolio.filters import PortfolioFilterSet
from wbportfolio.models import (
    Portfolio,
    PortfolioPortfolioThroughModel,
    PortfolioSynchronization,
    PriceComputation,
    TradeProposal,
)
from wbportfolio.models.portfolio import resynchronize_history_as_task
from wbportfolio.serializers import (
    ModelPortfolioModelSerializer,
    PortfolioModelSerializer,
    PortfolioPortfolioThroughModelSerializer,
    PortfolioRepresentationSerializer,
)

from .configs import (
    ModelPortfolioButtonConfig,
    ModelPortfolioDisplayConfig,
    ModelPortfolioEndpointConfig,
    ModelPortfolioTitleConfig,
    PortfolioButtonConfig,
    PortfolioDisplayConfig,
    PortfolioEndpointConfig,
    PortfolioPortfolioThroughModelDisplayConfig,
    PortfolioPortfolioThroughModelEndpointConfig,
    PortfolioPreviewConfig,
    PortfolioTitleConfig,
)
from .mixins import UserPortfolioRequestPermissionMixin


class PortfolioRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:portfolio"

    ordering_fields = ordering = search_fields = ("name",)
    queryset = Portfolio.objects.all()
    serializer_class = PortfolioRepresentationSerializer
    filterset_class = PortfolioFilterSet


class PortfolioModelViewSet(UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    filterset_class = PortfolioFilterSet
    serializer_class = PortfolioModelSerializer
    queryset = Portfolio.objects.all()

    search_fields = ("currency__key", "name")
    ordering_fields = search_fields
    ordering = ["name"]

    display_config_class = PortfolioDisplayConfig
    button_config_class = PortfolioButtonConfig
    title_config_class = PortfolioTitleConfig
    endpoint_config_class = PortfolioEndpointConfig
    preview_config_class = PortfolioPreviewConfig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related(
                "currency",
                "portfolio_synchronization",
            )
            .prefetch_related(
                "depends_on",
                "dependent_portfolios",
            )
        )

    @action(detail=True, methods=["PATCH"])
    def rebalance(self, request, pk=None):
        if (date_str := request.POST.get("trade_date", None)) and (
            model_portfolio_id := request.POST.get("model_portfolio", None)
        ):
            TradeProposal.objects.create(
                portfolio=Portfolio.objects.get(id=pk),
                model_portfolio=Portfolio.objects.get(id=model_portfolio_id),
                trade_date=datetime.strptime(date_str, "%Y-%m-%d"),
            )
            return Response({"send": True})
        raise HttpResponse("Bad Request", status=400)

    @action(detail=True, methods=["PATCH"])
    def resynchronize(self, request, pk=None):
        if request.user.has_perm("wbportfolio.administrate_instrument"):
            portfolio = get_object_or_404(Portfolio, pk=pk)
            instrument = request.GET.get("instrument", None)
            start, end = get_date_interval_from_request(request, request_type="POST")
            resynchronize_history_as_task.delay(portfolio.id, start, end, instrument_id=instrument)
            return Response({"send": True})
        return HttpResponse("Unauthorized", status=401)


class ModelPortfolioModelViewSet(PortfolioModelViewSet):
    serializer_class = ModelPortfolioModelSerializer
    endpoint_config_class = ModelPortfolioEndpointConfig
    display_config_class = ModelPortfolioDisplayConfig
    title_config_class = ModelPortfolioTitleConfig
    button_config_class = ModelPortfolioButtonConfig

    @action(detail=False, methods=["POST"])
    def createmodelportfolio(self, request, pk=None):
        if self.is_portfolio_manager:
            name = request.POST["name"]
            currency_id = request.POST["currency"]
            currency = Currency.objects.get(id=currency_id)
            portfolio_synchronization_id = request.POST.get("portfolio_synchronization", None)
            price_computation_id = request.POST.get("price_computation", None)
            portfolio_synchronization = (
                PortfolioSynchronization.objects.get(id=portfolio_synchronization_id)
                if portfolio_synchronization_id
                else None
            )
            create_index = request.POST.get("create_index", "false") == "true"
            index_parameters = {}
            if create_index:
                index_parameters["price_computation"] = (
                    PriceComputation.objects.get(id=price_computation_id) if price_computation_id else None
                )
            Portfolio.create_model_portfolio(
                name, currency, portfolio_synchronization=portfolio_synchronization, index_parameters=index_parameters
            )
            return Response({"send": True})
        raise HttpResponse("Unauthorized", status=403)

    def get_queryset(self):
        model_portfolios = PortfolioPortfolioThroughModel.objects.filter(
            type=PortfolioPortfolioThroughModel.Type.MODEL
        ).values("dependency_portfolio")
        return super().get_queryset().filter(id__in=model_portfolios)


class PortfolioPortfolioThroughModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    serializer_class = PortfolioPortfolioThroughModelSerializer
    queryset = PortfolioPortfolioThroughModel.objects.all()

    search_fields = ["dependency_portfolio__name"]

    display_config_class = PortfolioPortfolioThroughModelDisplayConfig
    endpoint_config_class = PortfolioPortfolioThroughModelEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(portfolio=self.kwargs["portfolio_id"])



from django.contrib import admin
from wbportfolio.models import PriceComputation

from .admin import SynchronizationAdmin


@admin.register(PriceComputation)
class PriceComputationModelAdmin(SynchronizationAdmin):
    fieldsets = (
        *SynchronizationAdmin.fieldsets,
        (
            "Synchronization",
            {
                "fields": (
                    "import_path",
                    "dependent_task",
                ),
                "classes": ("extrapretty", "wide"),
            },
        ),
    )



from dataclasses import asdict, dataclass, field, fields
from datetime import date as date_lib
from decimal import Decimal

import pandas as pd
from django.core.exceptions import ValidationError
from django.utils.functional import cached_property


@dataclass(frozen=True)
class Valuation:
    instrument: int
    net_value: Decimal
    outstanding_shares: Decimal = Decimal(0)


@dataclass(frozen=True)
class Position:
    underlying_instrument: int
    instrument_type: int
    weighting: Decimal
    currency: int
    date: date_lib

    asset_valuation_date: date_lib | None = None
    portfolio_created: int = None
    exchange: int = None
    is_estimated: bool = False
    country: int = None
    shares: Decimal | None = None
    is_cash: bool = False
    primary_classification: int = None
    favorite_classification: int = None
    market_capitalization_usd: float = None
    currency_fx_rate: float = 1
    market_share: float = None
    daily_liquidity: float = None
    volume_usd: float = None
    price: float = None

    def __add__(self, other):
        return Position(
            weighting=self.weighting + other.weighting,
            shares=self.shares + other.shares if (self.shares is not None and other.shares is not None) else None,
            **{f.name: getattr(self, f.name) for f in fields(Position) if f.name not in ["weighting", "shares"]},
        )


@dataclass(frozen=True)
class Portfolio:
    positions: tuple[Position]
    positions_map: dict[Position] = field(init=False, repr=False)

    def __post_init__(self):
        positions_map = {}
        for pos in self.positions:
            if pos.underlying_instrument in positions_map:
                positions_map[pos.underlying_instrument] += pos
            else:
                positions_map[pos.underlying_instrument] = pos
        object.__setattr__(self, "positions_map", positions_map)

    @cached_property
    def total_weight(self):
        return round(sum([pos.weighting for pos in self.positions]), 4)

    @cached_property
    def total_shares(self):
        return sum([pos.target_shares for pos in self.positions if pos.target_shares is not None])

    def to_df(self):
        return pd.DataFrame([asdict(pos) for pos in self.positions])

    def __len__(self):
        return len(self.positions)


@dataclass(frozen=True)
class Trade:
    underlying_instrument: int
    instrument_type: str
    currency: int
    date: date_lib

    effective_weight: Decimal
    target_weight: Decimal
    id: int | None = None
    effective_shares: Decimal = None
    target_shares: Decimal = None

    def __add__(self, other):
        return Trade(
            underlying_instrument=self.underlying_instrument,
            effective_weight=self.effective_weight + other.effective_weight,
            target_weight=self.target_weight + other.target_weight,
            target_shares=self.target_shares + other.target_shares
            if (self.target_shares is not None and other.target_shares is not None)
            else None,
            effective_shares=self.effective_shares + other.effective_shares
            if (self.effective_shares is not None and other.effective_shares is not None)
            else None,
            **{
                f.name: getattr(self, f.name)
                for f in fields(Trade)
                if f.name
                not in [
                    "effective_weight",
                    "target_weight",
                    "target_shares",
                    "effective_shares",
                    "underlying_instrument",
                ]
            },
        )

    @cached_property
    def delta_weight(self) -> Decimal:
        return self.target_weight - self.effective_weight

    def validate(self):
        if self.effective_weight < 0 or self.effective_weight > 1.0:
            raise ValidationError("Effective Weight needs to be in range [0, 1]")
        if self.target_weight < 0 or self.target_weight > 1.0:
            raise ValidationError("Target Weight needs to be in range [0, 1]")

    def normalize_target(self, total_target_weight: Decimal):
        t = Trade(
            target_weight=self.target_weight / total_target_weight if total_target_weight else self.target_weight,
            target_shares=self.target_shares / total_target_weight
            if (self.target_shares and total_target_weight)
            else self.target_shares,
            **{
                f.name: getattr(self, f.name)
                for f in fields(Trade)
                if f.name not in ["target_weight", "target_shares"]
            },
        )
        return t


@dataclass(frozen=True)
class TradeBatch:
    trades: tuple[Trade]
    trades_map: dict[Trade] = field(init=False, repr=False)

    def __post_init__(self):
        trade_map = {}
        for trade in self.trades:
            if trade.underlying_instrument in trade_map:
                trade_map[trade.underlying_instrument] += trade
            else:
                trade_map[trade.underlying_instrument] = trade
        object.__setattr__(self, "trades_map", trade_map)

    @cached_property
    def total_target_weight(self) -> Decimal:
        return round(sum([trade.target_weight for trade in self.trades]), 4)

    @cached_property
    def total_effective_weight(self) -> Decimal:
        return round(sum([trade.effective_weight for trade in self.trades]), 4)

    @cached_property
    def total_shares(self) -> Decimal:
        return sum([trade.target_shares for trade in self.trades if trade.target_shares is not None]) or Decimal(0)

    @cached_property
    def totat_abs_delta_weight(self) -> Decimal:
        return sum([abs(trade.delta_weight) for trade in self.trades])

    def __add__(self, other):
        return TradeBatch(tuple(self.trades + other.trades))

    def __len__(self):
        return len(self.trades)

    def validate(self):
        if float(self.total_target_weight) != 1.0:  # we do that to remove decimal over precision
            raise ValidationError(f"Total Weight cannot be different than 1 ({float(self.total_target_weight)})")

    def convert_to_portfolio(self):
        positions = []
        for instrument, trade in self.trades_map.items():
            positions.append(
                Position(
                    underlying_instrument=trade.underlying_instrument,
                    instrument_type=trade.instrument_type,
                    weighting=trade.target_weight,
                    shares=trade.target_shares,
                    currency=trade.currency,
                    date=trade.date,
                )
            )
        return Portfolio(tuple(positions))



from contextlib import suppress
from datetime import date, datetime, timedelta
from decimal import Decimal
from math import isclose
from typing import Any

import numpy as np
import pandas as pd
from celery import shared_task
from django.contrib.postgres.fields import DateRangeField
from django.db import models
from django.db.models import (
    BooleanField,
    Case,
    Exists,
    F,
    OuterRef,
    Q,
    QuerySet,
    Sum,
    Value,
    When,
)
from django.db.models.signals import post_save
from django.dispatch import receiver
from psycopg.types.range import DateRange
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.models import WBModel
from wbcore.utils.models import ActiveObjectManager, DeleteToDisableMixin
from wbfdm.contrib.metric.dispatch import compute_metrics
from wbfdm.models import Instrument
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.indexes import Index
from wbportfolio.models.portfolio_relationship import (
    InstrumentPortfolioThroughModel,
    PortfolioInstrumentPreferredClassificationThroughModel,
)
from wbportfolio.models.products import Product
from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from .utils import get_casted_portfolio_instrument


class DefaultPortfolioQueryset(QuerySet):
    def filter_invested_at_date(self, val_date: date) -> QuerySet:
        """
        Filter the queryset to get only portfolio invested at the given date
        """
        return self.filter(invested_timespan__startswith__lte=val_date, invested_timespan__endswith__gt=val_date)


class DefaultPortfolioManager(ActiveObjectManager):
    def get_queryset(self):
        return DefaultPortfolioQueryset(self.model).filter(is_active=True)

    def filter_invested_at_date(self, val_date: date):
        return self.get_queryset().filter_invested_at_date(val_date)


class ActiveTrackedPortfolioManager(DefaultPortfolioManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(asset_exists=Exists(AssetPosition.objects.filter(portfolio=OuterRef("pk"))))
            .filter(asset_exists=True, is_tracked=True)
        )


class PortfolioPortfolioThroughModel(models.Model):
    class Type(models.TextChoices):
        PRIMARY = "PRIMARY", "Primary"
        MODEL = "MODEL", "Model"
        BENCHMARK = "BENCHMARK", "Benchmark"
        INDEX = "INDEX", "Index"
        CUSTODIAN = "CUSTODIAN", "Custodian"

    portfolio = models.ForeignKey("wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependency_through")
    dependency_portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", on_delete=models.CASCADE, related_name="dependent_through"
    )
    type = models.CharField(choices=Type.choices, default=Type.PRIMARY, verbose_name="Type")

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["portfolio", "type"], name="unique_primary", condition=Q(type="PRIMARY")),
            models.UniqueConstraint(fields=["portfolio", "type"], name="unique_model", condition=Q(type="MODEL")),
        ]


class Portfolio(DeleteToDisableMixin, WBModel):
    assets: models.QuerySet[AssetPosition]

    name = models.CharField(
        max_length=255,
        verbose_name="Name",
        default="",
        help_text="The Name of the Portfolio",
    )

    currency = models.ForeignKey(
        to="currency.Currency",
        related_name="portfolios",
        on_delete=models.PROTECT,
        verbose_name="Currency",
        help_text="The currency of the portfolio.",
    )
    hedged_currency = models.ForeignKey(
        to="currency.Currency",
        related_name="hedged_portfolios",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name="Hedged Currency",
        help_text="The hedged currency of the portfolio.",
    )
    depends_on = models.ManyToManyField(
        "wbportfolio.Portfolio",
        symmetrical=False,
        related_name="dependent_portfolios",
        through="wbportfolio.PortfolioPortfolioThroughModel",
        through_fields=("portfolio", "dependency_portfolio"),
        blank=True,
        verbose_name="The portfolios this portfolio depends on",
    )

    portfolio_synchronization = models.ForeignKey(
        "wbportfolio.PortfolioSynchronization",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="portfolios",
        verbose_name="Portfolio Synchronization Method",
    )
    preferred_instrument_classifications = models.ManyToManyField(
        "wbfdm.Instrument",
        limit_choices_to=(models.Q(instrument_type__is_classifiable=True) & models.Q(level=0)),
        related_name="preferred_portfolio_classifications",
        through="wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel",
        through_fields=("portfolio", "instrument"),
        blank=True,
        verbose_name="The Preferred classification per instrument",
    )
    instruments = models.ManyToManyField(
        "wbfdm.Instrument",
        through=InstrumentPortfolioThroughModel,
        related_name="portfolios",
        blank=True,
        verbose_name="Instruments",
        help_text="Instruments linked to this instrument",
    )
    invested_timespan = DateRangeField(
        null=True, blank=True, help_text="Define when this portfolio is considered invested"
    )

    is_manageable = models.BooleanField(
        default=False,
        help_text="True if the portfolio can be manually modified (e.g. Trade proposal be submitted or total weight recomputed)",
    )
    is_tracked = models.BooleanField(
        default=True,
        help_text="True if the internal updating mechanism (e.g., Propagation, Synchronization etc...) needs to apply to this portfolio",
    )
    only_weighting = models.BooleanField(
        default=False,
        help_text="Indicates that this portfolio is only utilizing weights and disregards shares, e.g. a model portfolio",
    )

    last_synchronization = models.DateTimeField(blank=True, null=True, verbose_name="Last Synchronization Date")
    bank_accounts = models.ManyToManyField(
        to="directory.BankingContact",
        related_name="wbportfolio_portfolios",
        through="wbportfolio.PortfolioBankAccountThroughModel",
        blank=True,
    )
    objects = DefaultPortfolioManager()
    tracked_objects = ActiveTrackedPortfolioManager()

    @property
    def primary_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.PRIMARY
            ).dependency_portfolio

    @property
    def model_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.MODEL
            ).dependency_portfolio

    @property
    def benchmark_portfolio(self):
        with suppress(PortfolioPortfolioThroughModel.DoesNotExist):
            return PortfolioPortfolioThroughModel.objects.get(
                portfolio=self, type=PortfolioPortfolioThroughModel.Type.BENCHMARK
            ).dependency_portfolio

    @property
    def imported_assets(self):
        return self.assets.filter(is_estimated=False)

    def delete(self, **kwargs):
        super().delete(**kwargs)
        # We check if for all linked instruments, this portfolio was the last active one (if yes, we disable the instrument)
        if self.id:
            for instrument in self.instruments.iterator():
                if not instrument.portfolios.filter(is_active=True).exists():
                    instrument.delisted_date = date.today() - timedelta(days=1)
                    instrument.save()

    def _build_dto(self, val_date: date, **extra_filter_kwargs) -> PortfolioDTO:
        "returns the dto representation of this portfolio at the specified date"
        return PortfolioDTO(
            tuple([pos._build_dto() for pos in self.assets.filter(date=val_date, **extra_filter_kwargs)])
        )

    def is_invested_at_date(self, val_date: date) -> bool:
        return (
            self.invested_timespan
            and self.invested_timespan.upper > val_date
            and self.invested_timespan.lower <= val_date
        )

    def __str__(self):
        return f"{self.id:06} ({self.name})"

    class Meta:
        verbose_name = "Portfolio"
        verbose_name_plural = "Portfolios"

        notification_types = [
            create_notification_type(
                "wbportfolio.portfolio.check_custodian_portfolio",
                "Check Custodian Portfolio",
                "Sends a notification when a portfolio does not match with its custodian portfolio",
                True,
                True,
                True,
            ),
        ]

    @classmethod
    def create_model_portfolio(cls, name, currency, portfolio_synchronization=None, index_parameters=dict()):
        portfolio = cls.objects.create(
            is_manageable=True,
            name=name,
            currency=currency,
            portfolio_synchronization=portfolio_synchronization,
        )
        if index_parameters:
            index = Index.objects.create(name=name, currency=currency, **index_parameters)
            index.portfolios.all().delete()
            InstrumentPortfolioThroughModel.objects.update_or_create(
                instrument=index, defaults={"portfolio": portfolio}
            )
        return portfolio

    def is_active_at_date(self, val_date: date) -> bool:
        """
        Return if the base instrument has a total aum greater than 0
        :val_date: the date at which we need to evaluate if the portfolio is considered active
        """
        active_portfolio = self.is_active or self.deletion_datetime.date() > val_date
        if self.instruments.exists():
            return active_portfolio and any(
                [instrument.is_active_at_date(val_date) for instrument in self.instruments.all()]
            )
        return active_portfolio

    def get_aum(self, val_date: date) -> Decimal:
        """
        Return the total asset under management of the portfolio at the specified valuation date
        Args:
            val_date: The date at which aum needs to be computed
        Returns:
            The total AUM (0 if there is no position)
        """
        return self.assets.filter(date=val_date).aggregate(s=Sum("total_value_fx_portfolio"))["s"] or Decimal(0.0)

    def get_total_value(self, val_date):
        from wbportfolio.models.transactions.trades import Trade

        trades = Trade.valid_customer_trade_objects.filter(portfolio=self, transaction_date__lte=val_date)

        total_aum = Decimal(0)
        for underlying_instrument_id, sum_shares in (
            trades.values("underlying_instrument")
            .annotate(
                sum_shares=Sum("shares"),
            )
            .values_list("underlying_instrument", "sum_shares")
        ):
            with suppress(Instrument.DoesNotExist, InstrumentPrice.DoesNotExist):
                instrument = Instrument.objects.get(id=underlying_instrument_id)
                last_price = instrument.valuations.filter(date__lte=val_date).latest("date").net_value
                fx_rate = instrument.currency.convert(val_date, self.currency)
                total_aum += last_price * sum_shares * fx_rate
        return total_aum

    def _get_assets(self, with_estimated=True, with_cash=True):
        qs = self.assets
        if not with_estimated:
            qs = qs.filter(is_estimated=False)
        if not with_cash:
            qs = qs.exclude(underlying_instrument__is_cash=True)
        return qs

    def get_earliest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__gte=val_date)
        if qs.exists():
            return qs.earliest("date").date
        return None

    def get_latest_asset_position_date(self, val_date=None, with_estimated=False):
        qs = self._get_assets(with_estimated=with_estimated)
        if val_date:
            qs = qs.filter(date__lte=val_date)

        if qs.exists():
            return qs.latest("date").date
        return None

    # Asset Position Utility Functions
    def get_holding(self, val_date, exclude_cash=True, exclude_index=True):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date, weighting__gt=0)
        if exclude_index:
            qs = qs.exclude(underlying_security_instrument_type_key="index")
        return (
            qs.values("underlying_instrument__name")
            .annotate(total_value_fx_portfolio=Sum("total_value_fx_portfolio"), weighting=Sum("weighting"))
            .order_by("-total_value_fx_portfolio")
        )

    def _get_groupedby_df(
        self,
        group_by,
        val_date: date,
        exclude_cash: bool | None = False,
        exclude_index: bool | None = False,
        extra_filter_parameters: dict[str, Any] = None,
        **groupby_kwargs,
    ):
        qs = self._get_assets(with_cash=not exclude_cash).filter(date=val_date)
        if exclude_index:
            # We exclude only index that are not considered as cash. Setting exclude_cash to true convers this case.
            qs = qs.exclude(
                Q(underlying_security_instrument_type_key="index") & Q(underlying_instrument__is_cash=False)
            )
        if extra_filter_parameters:
            qs = qs.filter(**extra_filter_parameters)
        qs = group_by(qs, **groupby_kwargs).annotate(sum_weighting=Sum(F("weighting"))).order_by("-sum_weighting")
        df = pd.DataFrame(
            qs.values_list("aggregated_title", "sum_weighting"), columns=["aggregated_title", "weighting"]
        )
        if not df.empty:
            df.weighting = df.weighting.astype("float")
            df.weighting = df.weighting / df.weighting.sum()
            df = df.sort_values(by=["weighting"])
        return df.where(pd.notnull(df), None)

    def get_geographical_breakdown(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.country_group_by, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_currency_exposure(self, val_date, **kwargs):
        df = self._get_groupedby_df(AssetPosition.currency_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_market_cap_distribution(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.marketcap_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_security_instrument_type_key": "equity"},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_equity_liquidity(self, val_date, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.liquidity_group_by,
            val_date=val_date,
            exclude_cash=True,
            exclude_index=True,
            extra_filter_parameters={"underlying_security_instrument_type_key": "equity"},
            **kwargs,
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_industry_exposure(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(
            AssetPosition.group_by_primary, val_date=val_date, exclude_cash=True, exclude_index=True, **kwargs
        )
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_asset_allocation(self, val_date=None, **kwargs):
        df = self._get_groupedby_df(AssetPosition.cash_group_by, val_date=val_date, **kwargs)
        if not df.empty:
            df = df[df["weighting"] != 0]
        return df

    def get_adjusted_child_positions(self, val_date):
        if (
            child_positions := self.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        ).count() == 1:
            if portfolio := child_positions.first().underlying_instrument.primary_portfolio:
                child_positions = portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(date=val_date)
        for position in child_positions:
            if child_portfolio := position.underlying_instrument.primary_portfolio:
                index_positions = child_portfolio.assets.exclude(underlying_instrument__is_cash=True).filter(
                    date=val_date
                )

                for index_position in index_positions.all():
                    weighting = index_position.weighting * position.weighting
                    if weighting != 0:
                        yield {
                            "underlying_instrument_id": index_position.underlying_instrument.id,
                            "weighting": weighting,
                        }

    def get_longshort_distribution(self, val_date):
        df = pd.DataFrame(self.get_adjusted_child_positions(val_date))

        if not df.empty:
            df["is_cash"] = df.underlying_instrument_id.apply(lambda x: Instrument.objects.get(id=x).is_cash)
            df = df[~df["is_cash"]]
            df = (
                df[["underlying_instrument_id", "weighting"]].groupby("underlying_instrument_id").sum().astype("float")
            )
            df.weighting = df.weighting / df.weighting.sum()
            short_weight = df[df.weighting < 0].weighting.abs().sum()
            long_weight = df[df.weighting > 0].weighting.sum()
            total_weight = long_weight + short_weight
            return pd.DataFrame(
                [
                    {"title": "Long", "weighting": long_weight / total_weight},
                    {"title": "Short", "weighting": short_weight / total_weight},
                ]
            )
        return df

    def get_portfolio_contribution_df(self, start, end, with_cash=True, hedged_currency=None, only_equity=False):
        qs = self._get_assets(with_cash=with_cash).filter(date__gte=start, date__lte=end)
        if only_equity:
            qs = qs.filter(underlying_security_instrument_type_key="equity")
        return Portfolio.get_contribution_df(qs, hedged_currency=hedged_currency)

    def check_related_portfolio_at_date(self, val_date: date, related_portfolio: "Portfolio"):
        assets = AssetPosition.objects.filter(
            date=val_date, underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ).values("underlying_instrument__parent", "shares")
        assets1 = assets.filter(portfolio=self)
        assets2 = assets.filter(portfolio=related_portfolio)
        return assets1.difference(assets2)

    def change_at_date(
        self,
        val_date: date,
        recompute_weighting: bool = False,
        force_recompute_weighting: bool = False,
        synchronize: bool = True,
        **sync_kwargs,
    ):
        qs = (
            self.assets.filter(date=val_date)
            .filter(Q(total_value_fx_portfolio__isnull=False) | Q(weighting__isnull=False))
            .distinct()
        )

        # We normalize weight across the portfolio for a given date
        if (self.portfolio_synchronization or self.is_manageable or force_recompute_weighting) and qs.exists():
            total_weighting = qs.aggregate(s=Sum("weighting"))["s"]
            # We check if this actually necessary
            # (i.e. if the weight is already summed to 100%, it is already normalized)
            if not total_weighting or not isclose(total_weighting, Decimal(1.0), abs_tol=0.001) or recompute_weighting:
                total_value = qs.aggregate(s=Sum("total_value_fx_portfolio"))["s"]
                # TODO we change this because postgres doesn't support join statement in update (and total_value_fx_portfolio is a joined annoted field)
                for asset in qs:
                    if total_value:
                        asset.weighting = asset._total_value_fx_portfolio / total_value
                    elif total_weighting:
                        asset.weighting = asset.weighting / total_weighting
                    asset.save()
        if synchronize:
            for dependent_portfolio in self.dependent_portfolios.exclude(id=self.id).distinct():
                # Check if the dependent portfolio has a synchronization method and has assets at the specified date
                if (synchronization := dependent_portfolio.portfolio_synchronization) and (
                    dependent_portfolio.assets.filter(date__gte=val_date).exists()
                ):
                    # If this is true, we want to apply the synchronization at every synchronization period
                    # (scheduled crontab) from val_date to now.
                    if synchronization.propagate_history:
                        for _d in synchronization.dates_range(
                            val_date, dependent_portfolio.assets.latest("date").date, filter_daily=True
                        ):
                            synchronization.synchronize_as_task_si(
                                dependent_portfolio, _d, override_execution_datetime_validity=True
                            ).apply_async()
                    # Otherwise, we simply call a unique task for that date
                    else:
                        synchronization.synchronize_as_task_si(
                            dependent_portfolio, val_date, override_execution_datetime_validity=True
                        ).apply_async()

            # We check if there is an instrument attached to the portfolio with calculated NAV and price computation method
            for instrument in self.instruments.all():
                if price_computation := getattr(
                    get_casted_portfolio_instrument(instrument), "price_computation", None
                ):
                    inception_date = instrument.inception_date
                    if isinstance(inception_date, datetime):
                        inception_date = inception_date.date()

                    if isinstance(val_date, datetime):
                        val_date = val_date.date()

                    if inception_date is None or inception_date > val_date:
                        instrument.inception_date = val_date
                        instrument.save()
                    price_computation.compute(instrument, val_date, override_execution_datetime_validity=True)
            compute_metrics(val_date, basket=self)

    def propagate_or_update_assets(
        self,
        from_date: date,
        to_date: date,
        forward_price: bool | None = True,
        base_assets: dict[str, str] | None = None,
        delete_existing_assets: bool | None = False,
    ):
        # we don't propagate on already imported portfolio by default
        is_target_portfolio_imported = self.assets.filter(date=to_date, is_estimated=False).exists()
        if not base_assets:
            base_assets = dict()

        def _get_next_asset_valuation_date(current_asset_valuation_date):
            return (current_asset_valuation_date + pd.offsets.BDay(np.busday_count(from_date, to_date))).date()

        last_fx_date = CurrencyFXRates.objects.filter(date__lte=to_date).latest("date").date
        fx_rates = CurrencyFXRates.objects.filter(date=last_fx_date)
        assets = self.assets.filter(date=from_date)

        from_is_active = self.is_active_at_date(from_date)
        to_is_active = self.is_active_at_date(to_date)
        # # We check is the current assets are already stored and if there is no already stored valid assets
        # # With this, we ensure that we don't overwrite imported asset position with propagated ones.
        # assets_positions_next_day_count = self.assets.filter(date=to_date).count()
        if assets.exists() or base_assets:
            # Remove already existing assets
            if delete_existing_assets:
                self.assets.filter(date=to_date).delete()
            asset_list = list()
            # If base_assets is provided,
            # we assume that the portfolio composition is injected by this list of dictionary
            if base_assets:
                base_assets = (
                    base_assets
                    if isinstance(base_assets, dict)
                    else {asset_id: Decimal(1 / len(base_assets)) for asset_id in base_assets}
                )

            remaining_base_assets = base_assets.copy()
            # Loop over existing assets and construct the propagation assets list
            for asset in assets.all():
                # if a composition is provided, we ensure that existing assets don't deviate from it
                if (base_assets and asset.underlying_instrument.id in base_assets.keys()) or not base_assets:
                    next_asset_valuation_date = _get_next_asset_valuation_date(asset.asset_valuation_date)
                    with suppress(ValueError):
                        asset_list.append(
                            {
                                "initial_price": (
                                    asset.initial_price
                                    if asset._price is not None
                                    else asset.underlying_instrument.get_price(from_date)
                                ),
                                "asset_valuation_date": next_asset_valuation_date,
                                "weighting": asset.weighting,
                                "next_initial_price": asset.underlying_instrument.get_price(next_asset_valuation_date),
                                "underlying_instrument": asset.underlying_instrument,
                                "exchange": asset.exchange,
                                "portfolio": asset.portfolio,
                                "portfolio_created": asset.portfolio_created,
                                "currency": asset.currency,
                                "initial_shares": asset.initial_shares,
                            }
                        )
                remaining_base_assets.pop(asset.underlying_instrument.id, None)
            # We ensure that the propagation assets list contains the proposed composition
            for asset_id, weighting in remaining_base_assets.items():
                instrument = Instrument.objects.get(id=asset_id)
                with suppress(ValueError):
                    asset_list.append(
                        {
                            "underlying_instrument": instrument,
                            "initial_price": instrument.get_price(from_date),
                            "next_initial_price": instrument.get_price(to_date),
                            "asset_valuation_date": to_date,
                            "initial_shares": None,
                            "portfolio": self,
                            "currency": instrument.currency,
                            "weighting": weighting,
                        }
                    )

            df = pd.DataFrame(asset_list)
            if not df.empty:
                df[["initial_price", "weighting", "next_initial_price"]] = df[
                    ["initial_price", "weighting", "next_initial_price"]
                ].astype("float")
                idxx = pd.isnull(df["initial_price"]) & ~pd.isnull(df["next_initial_price"])
                df.loc[idxx, "initial_price"] = df.loc[idxx, "next_initial_price"]
                if forward_price:
                    idx = pd.isnull(df["next_initial_price"])
                    df.loc[idx, "next_initial_price"] = df.loc[idx, "initial_price"]
                df = df.dropna(axis=0, subset=["next_initial_price", "initial_price"])
                # Normalize weight to 100%. Exclude portfolio were sum of weight equals 0 (e.g. short/long portfolio)
                if df.weighting.sum() != 0:
                    df["weighting"] /= df.weighting.sum()
                df.loc[:, "perf"] = df.loc[:, "next_initial_price"] / df.loc[:, "initial_price"]
                df["contribution"] = df.perf * df.weighting
                df.loc[:, "next_weighting"] = df.contribution

                if df.contribution.sum() != 0:
                    df.loc[:, "next_weighting"] /= df.contribution.sum()

                # Normalize next weighting
                if df.next_weighting.sum() != 0:
                    df.next_weighting /= df.next_weighting.sum()
                df = df.replace([np.inf, -np.inf, np.nan], None)
                df.loc[(df["next_weighting"] < -1) | (df["next_weighting"] > 1), "next_weighting"] = df.loc[
                    (df["next_weighting"] < -1) | (df["next_weighting"] > 1), "weighting"
                ]  # if the next weighting is not including within -1 and 1 range, we default to the initial weighting
                if not df.empty:
                    for row in df.to_dict("records"):
                        weighting = Decimal(row["next_weighting"]) if row["next_weighting"] else row["weighting"]
                        if from_is_active and not to_is_active:
                            weighting = Decimal(0.0)
                        try:
                            initial_currency_fx_rate = (
                                fx_rates.get(currency=self.currency).value
                                / fx_rates.get(currency=row["currency"]).value
                            )
                        except CurrencyFXRates.DoesNotExist:
                            initial_currency_fx_rate = Decimal(1)
                        defaults = {
                            "initial_currency_fx_rate": initial_currency_fx_rate,
                            "weighting": weighting,
                            "initial_price": Decimal(row["next_initial_price"]),
                            "initial_shares": row["initial_shares"],
                            "asset_valuation_date": row["asset_valuation_date"],
                            "is_estimated": True,
                        }
                        get_parameters = {
                            "underlying_instrument": row["underlying_instrument"],
                            "portfolio": self,
                            "currency": row["currency"],
                            "date": to_date,
                        }
                        if exchange := row.get("exchange", None):
                            get_parameters["exchange"] = exchange
                        if portfolio_created := row.get("portfolio_created", None):
                            get_parameters["portfolio_created"] = portfolio_created
                        # We check if an asset position already exists and if so, if it is estimated
                        # (otherwise we don't propagate it)
                        if _asset := AssetPosition.objects.filter(**get_parameters).first():
                            _asset.underlying_instrument_price = None  # we unset the previously linked underlying instrument price in case it was linked to the wrong underlying price (e.g too early)
                            if not from_is_active and not to_is_active:
                                _asset.delete()
                            elif not is_target_portfolio_imported and _asset.is_estimated:
                                for k, v in defaults.items():
                                    setattr(_asset, k, v)
                                _asset.save()
                        elif from_is_active and to_is_active and not is_target_portfolio_imported:
                            AssetPosition.objects.create(**get_parameters, **defaults)

    def import_positions_at_date(self, portfolio: PortfolioDTO, val_date: date, post_processing: bool = False):
        if not portfolio:
            return
        left_over_positions = self.assets.filter(date=val_date)

        # We convert the positions into a dataframe in order to handle positions that are considered duplicates
        # In that case, we sum up fields such as weighting and shares.
        # Position are assumed serialized otherwise the groupby on dataframe can't handle django object
        index_columns = ["portfolio_id", "date", "underlying_instrument_id", "portfolio_created_id"]
        float_columns = [
            "weighting",
            "initial_currency_fx_rate",
            "initial_shares",
            "initial_price",
        ]
        df = portfolio.to_df().rename(
            columns={
                "currency_fx_rate": "initial_currency_fx_rate",
                "shares": "initial_shares",
                "price": "initial_price",
                "currency": "currency_id",
                "underlying_instrument": "underlying_instrument_id",
                "portfolio_created": "portfolio_created_id",
                "exchange": "exchange_id",
            }
        )
        df["portfolio_id"] = self.id
        df = df[index_columns + float_columns + ["is_estimated", "currency_id"]]
        df[float_columns] = df[float_columns].astype("float")
        df = df.groupby(index_columns, as_index=False, dropna=False).agg(
            {
                **{field: "first" for field in df.columns.difference(index_columns + float_columns)},
                "weighting": "sum",
                "initial_shares": "sum",
                "initial_currency_fx_rate": "mean",
                "initial_price": "mean",
            }
        )
        df = df.replace([np.inf, -np.inf, np.nan], None)

        for position in df.to_dict("records"):
            obj, _ = AssetPosition.unannotated_objects.update_or_create(
                portfolio_id=position["portfolio_id"],
                date=position["date"],
                underlying_instrument_id=position["underlying_instrument_id"],
                portfolio_created_id=position["portfolio_created_id"],
                defaults=position,
            )
            left_over_positions = left_over_positions.exclude(id=obj.id)
        left_over_positions.delete()
        if post_processing:
            trigger_portfolio_change_as_task.delay(self.id, val_date)

    def resynchronize_history(self, from_date: date, to_date: date, instrument: Instrument | None = None):
        if (synchronisation_method := self.portfolio_synchronization) and self.assets.exists():
            if not from_date:
                from_date = self.assets.earliest("date").date
            if not to_date:
                to_date = self.assets.latest("date").date
            # loop over every week day and trigger synchronization task in order
            if to_date <= from_date:
                raise ValueError("bound needs to be valid")
            for sync_datetime in synchronisation_method.dates_range(from_date, to_date, filter_daily=True):
                synchronisation_method.synchronize(
                    self, sync_datetime.date(), override_execution_datetime_validity=True
                )
        if instrument:
            price_computation_method = None
            try:
                price_computation_method = Product.objects.get(id=instrument.id).price_computation
            except Product.DoesNotExist:
                with suppress(Index.DoesNotExist):
                    price_computation_method = Index.objects.get(id=instrument.id).price_computation
            if price_computation_method and instrument.prices.exists():
                if to_date <= from_date:
                    raise ValueError("bound needs to be valid")
                if not from_date:
                    from_date = instrument.prices.earliest("date").date
                if not to_date:
                    to_date = instrument.prices.latest("date").date
                # loop over every week day and trigger synchronization task in order
                for sync_datetime in price_computation_method.dates_range(from_date, to_date, filter_daily=True):
                    price_computation_method.compute(
                        instrument, sync_datetime.date(), override_execution_datetime_validity=True
                    )

    def update_preferred_classification_per_instrument(self):
        # Function to automatically assign Preferred instrument based on the assets' underlying instruments of the
        # attached wbportfolio
        instruments = filter(
            None,
            map(
                lambda x: Instrument.objects.get(id=x["underlying_instrument"]).get_classifable_ancestor(
                    include_self=True
                ),
                self.assets.values("underlying_instrument").distinct("underlying_instrument"),
            ),
        )
        leftovers_instruments = list(
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(portfolio=self).values_list(
                "instrument", flat=True
            )
        )
        for instrument in instruments:
            other_classifications = instrument.classifications.filter(group__is_primary=False)
            default_classification = None
            if other_classifications.count() == 1:
                default_classification = other_classifications.first()
            if not PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument
            ).exists():
                PortfolioInstrumentPreferredClassificationThroughModel.objects.create(
                    portfolio=self,
                    instrument=instrument,
                    classification=default_classification,
                    classification_group=default_classification.group if default_classification else None,
                )
            if instrument.id in leftovers_instruments:
                leftovers_instruments.remove(instrument.id)

        for instrument_id in leftovers_instruments:
            PortfolioInstrumentPreferredClassificationThroughModel.objects.filter(
                portfolio=self, instrument=instrument_id
            ).delete()

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfolio"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:portfoliorepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    @classmethod
    def _get_or_create_portfolio(cls, instrument_handler, portfolio_data):
        if isinstance(portfolio_data, int):
            return Portfolio.objects.get(id=portfolio_data)
        instrument = portfolio_data
        if isinstance(portfolio_data, dict):
            instrument = instrument_handler.process_object(instrument, only_security=False, read_only=True)[0]
        return instrument.primary_portfolio

    def check_share_diff(self, val_date: date) -> bool:
        return self.assets.filter(Q(date=val_date) & ~Q(initial_shares=F("initial_shares_at_custodian"))).exists()

    @classmethod
    def get_contribution_df(
        cls,
        qs,
        need_normalize=False,
        groupby_label_id="underlying_security",
        groubpy_label_title="underlying_instrument__name_repr",
        currency_fx_rate_label="currency_fx_rate",
        hedged_currency=None,
    ):
        # qs = AssetPosition.annotate_underlying_instrument(qs)
        weight_label = "weighting" if not need_normalize else "total_value_fx_portfolio"
        qs = qs.annotate(
            is_hedged=Case(
                When(
                    underlying_instrument__currency__isnull=False,
                    underlying_instrument__currency=hedged_currency,
                    then=Value(True),
                ),
                default=Value(False),
                output_field=BooleanField(),
            ),
            coalesce_currency_fx_rate=Case(
                When(is_hedged=True, then=Value(Decimal(1.0))),
                default=F(currency_fx_rate_label),
                output_field=models.BooleanField(),
            ),
        ).select_related("underlying_instrument")
        df = pd.DataFrame(
            qs.values(
                "date",
                "price",
                "coalesce_currency_fx_rate",
                groupby_label_id,
                groubpy_label_title,
                weight_label,
            ),
            columns=[
                "date",
                "price",
                "coalesce_currency_fx_rate",
                groupby_label_id,
                groubpy_label_title,
                weight_label,
            ],
        )
        if not df.empty:
            df = df[df[weight_label] != 0]
            df.date = pd.to_datetime(df.date)
            df["price_fx_portfolio"] = df.price * df.coalesce_currency_fx_rate

            df[["price", "price_fx_portfolio", weight_label, "coalesce_currency_fx_rate"]] = df[
                ["price", "price_fx_portfolio", weight_label, "coalesce_currency_fx_rate"]
            ].astype("float")

            df[groupby_label_id] = df[groupby_label_id].fillna(0)
            df[groubpy_label_title] = df[groubpy_label_title].fillna("N/A")
            df_static = df[[groupby_label_id, groubpy_label_title]].groupby(groupby_label_id, dropna=False).first()

            df = (
                df[
                    [
                        groupby_label_id,
                        "date",
                        "price",
                        "price_fx_portfolio",
                        weight_label,
                        "coalesce_currency_fx_rate",
                    ]
                ]
                .groupby(["date", groupby_label_id], dropna=False)
                .agg(
                    {
                        "price": "mean",
                        "price_fx_portfolio": "mean",
                        weight_label: "sum",
                        "coalesce_currency_fx_rate": "mean",
                    }
                )
                .reset_index()
                .set_index("date")
                .sort_index()
            )
            df[weight_label] = df[weight_label].fillna(0)
            value = df.pivot_table(
                index="date",
                columns=[groupby_label_id],
                values=weight_label,
                fill_value=0,
                aggfunc="sum",
            )
            weights_ = value
            if need_normalize:
                total_value_price = df[weight_label].groupby("date", dropna=False).sum()
                weights_ = value.divide(total_value_price, axis=0)
            prices_usd = (
                df.pivot_table(
                    index="date",
                    columns=[groupby_label_id],
                    values="price_fx_portfolio",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            rates_fx = (
                df.pivot_table(
                    index="date",
                    columns=[groupby_label_id],
                    values="coalesce_currency_fx_rate",
                    aggfunc="mean",
                )
                .replace(0, np.nan)
                .bfill()
            )

            prices_usd = prices_usd.ffill()
            performance_prices = prices_usd / prices_usd.shift(1, axis=0) - 1
            contributions_prices = performance_prices.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_prices = (1 + contributions_prices.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_prices = contributions_prices.multiply(total_contrib_prices, axis=0).sum(skipna=False)
            monthly_perf_prices = (1 + performance_prices).dropna(how="all").product(axis=0, skipna=False) - 1

            rates_fx = rates_fx.ffill()
            performance_rates_fx = rates_fx / rates_fx.shift(1, axis=0) - 1
            contributions_rates_fx = performance_rates_fx.multiply(weights_.shift(1, axis=0)).dropna(how="all")
            total_contrib_rates_fx = (1 + contributions_rates_fx.sum(axis=1)).shift(1, fill_value=1.0).cumprod()
            contributions_rates_fx = contributions_rates_fx.multiply(total_contrib_rates_fx, axis=0).sum(skipna=False)
            monthly_perf_rates_fx = (1 + performance_rates_fx).dropna(how="all").product(axis=0, skipna=False) - 1

            res = pd.concat(
                [
                    df_static,
                    monthly_perf_prices,
                    monthly_perf_rates_fx,
                    contributions_prices,
                    contributions_rates_fx,
                    weights_.iloc[0, :],
                    weights_.iloc[-1, :],
                    value.iloc[0, :],
                    value.iloc[-1, :],
                ],
                axis=1,
            ).reset_index()
            res.columns = [
                groupby_label_id,
                groubpy_label_title,
                "performance_total",
                "performance_forex",
                "contribution_total",
                "contribution_forex",
                "allocation_start",
                "allocation_end",
                "total_value_start",
                "total_value_end",
            ]

            return res.replace([np.inf, -np.inf, np.nan], 0)
        return pd.DataFrame()


@receiver(post_save, sender="wbportfolio.Product")
@receiver(post_save, sender="wbportfolio.ProductGroup")
@receiver(post_save, sender="wbportfolio.Index")
def post_product_creation(sender, instance, created, raw, **kwargs):
    if not raw and (created or not InstrumentPortfolioThroughModel.objects.filter(instrument=instance).exists()):
        portfolio = Portfolio.objects.create(
            name=f"Portfolio: {instance.name}",
            currency=instance.currency,
            invested_timespan=DateRange(instance.inception_date if instance.inception_date else date.min, date.max),
        )
        InstrumentPortfolioThroughModel.objects.get_or_create(instrument=instance, defaults={"portfolio": portfolio})


@shared_task(queue="portfolio")
def resynchronize_history_as_task(portfolio_id: int, from_date: date, to_date: date, instrument_id: int | None = None):
    portfolio = Portfolio.objects.get(id=portfolio_id)
    instrument = Instrument.objects.get(id=instrument_id) if instrument_id else None
    portfolio.resynchronize_history(from_date, to_date, instrument=instrument)


@shared_task(queue="portfolio")
def trigger_portfolio_change_as_task(portfolio_id, val_date, **kwargs):
    portfolio = Portfolio.all_objects.get(id=portfolio_id)
    portfolio.change_at_date(val_date, **kwargs)


@shared_task(queue="portfolio")
def propagate_or_update_portfolio_assets_as_task(portfolio_id, from_date, to_date, **kwargs):
    portfolio = Portfolio.objects.get(id=portfolio_id)
    portfolio.propagate_or_update_assets(from_date, to_date, **kwargs)



from datetime import date
from decimal import Decimal
from unittest.mock import patch

import pytest
from django.db.models import F, Sum
from django.forms.models import model_to_dict
from faker import Faker
from pandas.tseries.offsets import BDay
from psycopg.types.range import DateRange
from wbcore.contrib.geography.factories import CountryFactory
from wbportfolio.models import (
    AssetPosition,
    Portfolio,
    PortfolioInstrumentPreferredClassificationThroughModel,
    PortfolioSynchronization,
)
from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from .test_synchronization import callback
from .utils import PortfolioTestMixin

fake = Faker()


@pytest.mark.django_db
class TestPortfolioModel(PortfolioTestMixin):
    def test_init(self, portfolio):
        assert portfolio.id is not None

    def test_str(self, portfolio):
        assert str(portfolio) == f"{portfolio.id:06} ({portfolio.name})"

    def test_get_assets(self, portfolio, product, cash, asset_position_factory):
        asset_position_factory.create_batch(4, portfolio=portfolio, underlying_instrument=product)
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=cash)
        assert portfolio._get_assets().count() == 5
        assert portfolio._get_assets(with_cash=False).count() == 4

    def test_get_earliest_asset_position_date(self, portfolio, asset_position_factory):
        asset_position_factory.create_batch(5, portfolio=portfolio)
        assert portfolio.get_earliest_asset_position_date() == AssetPosition.objects.earliest("date").date

    def test_get_latest_asset_position_date(self, portfolio, asset_position_factory):
        asset_position_factory.create_batch(5, portfolio=portfolio)
        assert portfolio.get_latest_asset_position_date() == AssetPosition.objects.latest("date").date

    def test_get_holding(self, portfolio_factory, asset_position_factory, equity, weekday):
        portfolio = portfolio_factory.create()
        asset_position_factory.create(portfolio=portfolio, date=weekday, initial_price=1, initial_shares=10)
        a2 = asset_position_factory.create(
            portfolio=portfolio,
            date=weekday,
            initial_price=1,
            initial_shares=40,
            underlying_instrument=equity,
            portfolio_created=portfolio_factory.create(),
        )
        asset_position_factory.create(portfolio=portfolio, date=weekday, initial_price=1, initial_shares=50)
        a4 = asset_position_factory.create(
            portfolio=portfolio,
            date=weekday,
            initial_price=1,
            initial_shares=30,
            underlying_instrument=equity,
            portfolio_created=portfolio_factory.create(),
        )
        assert (
            portfolio.get_holding(weekday)
            .filter(underlying_instrument=equity)
            .values_list("total_value_fx_portfolio", flat=True)[0]
            == a2._total_value_fx_portfolio + a4._total_value_fx_portfolio
        )

    def test_get_groupeby(self, portfolio, asset_position_factory, weekday):
        a1 = asset_position_factory.create(
            portfolio=portfolio, date=weekday, initial_price=1, initial_shares=10, weighting=0.1
        )
        asset_position_factory.create(
            portfolio=portfolio, date=weekday, initial_price=1, initial_shares=40, weighting=0.4
        )
        asset_position_factory.create(
            portfolio=portfolio, date=weekday, initial_price=1, initial_shares=50, weighting=0.5
        )

        def groupby(qs, **kwargs):
            return qs.annotate(aggregated_title=F("underlying_instrument__ticker"))

        df = portfolio._get_groupedby_df(groupby, weekday)
        assert df.aggregated_title[2] == a1.underlying_instrument.ticker

    def test_get_geographical_breakdown(self, portfolio, asset_position_factory, equity_factory, weekday):
        c1 = CountryFactory.create()
        c2 = CountryFactory.create()
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(country=c1), date=weekday
        )
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(country=c1), date=weekday
        )
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(country=c2), date=weekday
        )
        assert portfolio.get_geographical_breakdown(weekday).shape[0] == 2

    def test_get_currency_exposure(self, portfolio, asset_position_factory, currency_factory, equity_factory, weekday):
        a1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=equity_factory.create(),
            currency=currency_factory.create(),
            date=weekday,
        )
        asset_position_factory.create(
            portfolio=portfolio, underlying_instrument=equity_factory.create(), currency=a1.currency, date=weekday
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=equity_factory.create(),
            currency=currency_factory.create(),
            date=weekday,
        )
        assert portfolio.get_currency_exposure(weekday).shape[0] == 2

    def test_get_equity_market_cap_distribution(
        self, portfolio, equity, asset_position_factory, instrument_price_factory
    ):
        price = instrument_price_factory.create(instrument=equity)
        asset_position_factory.create_batch(10, portfolio=portfolio, date=price.date, underlying_instrument=equity)

        assert not portfolio.get_equity_market_cap_distribution(price.date).empty

    def test_get_get_equity_liquidity(self, portfolio, asset_position_factory, equity_factory, weekday):
        asset_position_factory.create_batch(
            10, portfolio=portfolio, date=weekday, underlying_instrument=equity_factory.create()
        )
        assert not portfolio.get_equity_liquidity(weekday).empty

    def test_get_industry_exposure(
        self, portfolio, asset_position_factory, weekday, classification_group, classification_factory, equity_factory
    ):
        parent_classification = classification_factory.create(group=classification_group)
        asset_position_factory.create_batch(
            10,
            portfolio=portfolio,
            date=weekday,
            underlying_instrument=equity_factory.create(
                classifications=[
                    classification_factory.create(group=classification_group, parent=parent_classification)
                ]
            ),
        )
        assert not portfolio.get_industry_exposure(weekday).empty

    def test_get_asset_allocation(self, portfolio, equity, cash, index_factory, asset_position_factory, weekday):
        index = index_factory.create(is_cash=True)
        asset_position_factory.create_batch(10, portfolio=portfolio, date=weekday, underlying_instrument=equity)
        asset_position_factory.create(portfolio=portfolio, date=weekday, underlying_instrument=index)
        asset_position_factory.create(portfolio=portfolio, date=weekday, underlying_instrument=cash)
        assert portfolio.get_asset_allocation(weekday).shape[0] == 2

    def test_get_portfolio_contribution_df(self, portfolio, asset_position_factory, instrument_factory, weekday):
        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        end = (weekday + BDay(1)).date()
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i1,
            date=weekday,
            initial_price=100,
            initial_shares=10,
            initial_currency_fx_rate=1,
            weighting=0.25,
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i2,
            date=weekday,
            initial_price=100,
            initial_shares=30,
            initial_currency_fx_rate=1,
            weighting=0.75,
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i1,
            date=end,
            initial_price=120,
            initial_shares=10,
            initial_currency_fx_rate=1,
            weighting=0.33,
        )
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i2,
            date=end,
            initial_price=80,
            initial_shares=30,
            initial_currency_fx_rate=1,
            weighting=0.66,
        )
        res = portfolio.get_portfolio_contribution_df(weekday, end)
        assert 0.05 == pytest.approx(res.contribution_total[0])
        assert -0.15 == pytest.approx(res.contribution_total[1])
        assert 0.2 == pytest.approx(res.performance_total[0])
        assert -0.2 == pytest.approx(res.performance_total[1])

    def test_get_longshort_distribution(
        self, asset_position_factory, portfolio_factory, cash, index_factory, equity_factory, weekday
    ):
        portfolio = portfolio_factory.create()

        ind1 = index_factory.create(is_cash=False)
        short_underlying_portfolio = ind1.portfolio

        ind2 = index_factory.create(is_cash=False)
        long_underlying_portfolio = ind2.portfolio

        asset_position_factory.create(date=weekday, portfolio=portfolio, weighting=-1.0, underlying_instrument=ind1)
        short_p1 = asset_position_factory.create(
            date=weekday, portfolio=short_underlying_portfolio, underlying_instrument=cash
        )
        short_p2 = asset_position_factory.create(
            date=weekday,
            portfolio=short_underlying_portfolio,
            underlying_instrument=equity_factory.create(is_cash=False),
            weighting=1 - short_p1.weighting,
        )

        asset_position_factory.create(date=weekday, portfolio=portfolio, weighting=1.0, underlying_instrument=ind2)
        long_p1 = asset_position_factory.create(
            date=weekday,
            portfolio=long_underlying_portfolio,
            underlying_instrument=equity_factory.create(is_cash=False),
        )
        long_p2 = asset_position_factory.create(
            date=weekday,
            portfolio=long_underlying_portfolio,
            underlying_instrument=equity_factory.create(is_cash=False),
            weighting=1 - long_p1.weighting,
        )

        res = portfolio.get_longshort_distribution(weekday)
        total_weight = abs(short_p2.weighting) + long_p1.weighting + long_p2.weighting
        assert Decimal(res.weighting[0]) == pytest.approx(
            (long_p1.weighting + long_p2.weighting) / total_weight, rel=Decimal(1e-4)
        )
        assert Decimal(res.weighting[1]) == pytest.approx((abs(short_p2.weighting)) / total_weight, rel=Decimal(1e-4))

    def test_change_at_date(self, asset_position_factory, portfolio, weekday):
        asset_position_factory.create_batch(10, portfolio=portfolio, date=weekday)

        portfolio.change_at_date(weekday)
        total_value = AssetPosition.objects.aggregate(s=Sum("total_value_fx_portfolio"))["s"]
        for pos in AssetPosition.objects.all():
            assert float(pos.weighting) == pytest.approx(float(pos.total_value_fx_portfolio / total_value), rel=1e-2)

    @patch.object(PortfolioSynchronization, "synchronize_as_task_si")
    def test_change_at_date_with_dependent_portfolio(
        self, mock_synchronize, asset_position_factory, portfolio_factory, portfolio_synchronization, weekday
    ):
        mock_synchronize.return_value = callback
        base_portfolio = portfolio_factory.create()
        asset_position_factory.create_batch(10, portfolio=base_portfolio, date=weekday)

        dependent_portfolio1 = portfolio_factory.create(portfolio_synchronization=portfolio_synchronization)
        dependent_portfolio2 = portfolio_factory.create(portfolio_synchronization=portfolio_synchronization)
        asset_position_factory.create_batch(10, portfolio=dependent_portfolio1, date=weekday)
        asset_position_factory.create_batch(10, portfolio=dependent_portfolio2, date=weekday)

        dependent_portfolio1.depends_on.add(base_portfolio)
        dependent_portfolio2.depends_on.add(base_portfolio)
        base_portfolio.change_at_date(weekday)

        assert mock_synchronize.call_count == 2

    @patch.object(PortfolioSynchronization, "synchronize_as_task_si")
    @pytest.mark.parametrize("portfolio_synchronization__propagate_history", [True])
    def test_change_at_date_with_dependent_portfolio_and_history(
        self, mock_synchronize, portfolio_synchronization, asset_position_factory, portfolio_factory, weekday
    ):
        mock_synchronize.return_value = callback
        base_portfolio = portfolio_factory.create()
        asset_position_factory.create_batch(10, portfolio=base_portfolio, date=weekday)

        dependent_portfolio1 = portfolio_factory.create(portfolio_synchronization=portfolio_synchronization)
        dependent_portfolio2 = portfolio_factory.create(portfolio_synchronization=portfolio_synchronization)
        asset_position_factory.create_batch(10, portfolio=dependent_portfolio1, date=weekday)
        asset_position_factory.create_batch(10, portfolio=dependent_portfolio2, date=weekday)
        asset_position_factory.create_batch(10, portfolio=dependent_portfolio1, date=weekday + BDay(6))
        asset_position_factory.create_batch(10, portfolio=dependent_portfolio2, date=weekday + BDay(6))
        total_nb_days = ((weekday + BDay(6)).date() - weekday).days
        dependent_portfolio1.depends_on.add(base_portfolio)
        dependent_portfolio2.depends_on.add(base_portfolio)
        base_portfolio.change_at_date(weekday)

        assert mock_synchronize.call_count == 2 * (total_nb_days + 1)

    def test_is_active_at_date(
        self,
        portfolio,
        instrument_factory,
    ):
        # a portfolio is active at a date if it is active or the deletion time is greater than that date AND if there is instruments attached, at least one instrument is still active as well

        assert portfolio.is_active
        assert portfolio.is_active_at_date(fake.date_object())

        portfolio.delete()  # soft deletion
        assert portfolio.is_active_at_date(fake.past_date())
        assert not portfolio.is_active_at_date(fake.future_date())

    def test_is_active_at_date_with_instruments(
        self,
        portfolio,
        instrument_factory,
    ):
        i1 = instrument_factory.create(inception_date=date.today(), delisted_date=None)
        i2 = instrument_factory.create(inception_date=date.today(), delisted_date=None)
        portfolio.instruments.add(i1)
        portfolio.instruments.add(i2)

        assert i1.is_active_at_date(fake.future_date())
        assert i2.is_active_at_date(fake.future_date())
        assert portfolio.is_active_at_date(fake.future_date())

        i1.delisted_date = date.today()
        i1.save()
        assert portfolio.is_active_at_date(fake.future_date())

        i2.delisted_date = date.today()
        i2.save()
        assert not portfolio.is_active_at_date(
            fake.date_object()
        )  # as no instrument is active, even if the portfolio is active at any date, the portfolio is consiodered inactive

    def test_propagate_or_update_assets(
        self, portfolio, asset_position_factory, instrument_factory, instrument_price_factory, weekday
    ):
        next_day = (weekday + BDay(1)).date()

        i1 = instrument_factory.create(currency=portfolio.currency)
        price1_1 = instrument_price_factory.create(instrument=i1, date=weekday)
        price1_2 = instrument_price_factory.create(instrument=i1, date=next_day)
        i2 = instrument_factory.create(currency=portfolio.currency)
        price2_1 = instrument_price_factory.create(instrument=i2, date=weekday)
        price2_2 = instrument_price_factory.create(instrument=i2, date=next_day)
        i3 = instrument_factory.create(currency=portfolio.currency)
        price3_1 = instrument_price_factory.create(instrument=i3, date=weekday)
        price3_2 = instrument_price_factory.create(instrument=i3, date=next_day)
        i4 = instrument_factory.create(currency=portfolio.currency)

        a1_1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i1,
            underlying_instrument_price=price1_1,
            date=weekday,
            weighting=Decimal(0.4),
        )
        a2_1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i2,
            underlying_instrument_price=price2_1,
            date=weekday,
            weighting=Decimal(0.3),
        )
        a3_1 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_instrument=i3,
            underlying_instrument_price=price3_1,
            date=weekday,
            weighting=Decimal(0.2),
        )
        a4_1 = asset_position_factory.create(  # noqa
            portfolio=portfolio,
            underlying_instrument=i4,
            underlying_instrument_price=None,  # the price won't be created automatically by the fixture, we expect this position to be removed from the propagated portfolio
            date=weekday,
            weighting=Decimal(0.1),
        )

        # Test basic output
        portfolio.propagate_or_update_assets(weekday, next_day, delete_existing_assets=False)
        a1_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i1)
        a2_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i2)
        a3_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i3)
        with pytest.raises(AssetPosition.DoesNotExist):
            AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i4)

        assert a1_2.initial_price == pytest.approx(price1_2.net_value, rel=Decimal(1e-4))
        assert a2_2.initial_price == pytest.approx(price2_2.net_value, rel=Decimal(1e-4))
        assert a3_2.initial_price == pytest.approx(price3_2.net_value, rel=Decimal(1e-4))

        contrib_1 = a1_1.weighting * price1_2.net_value / price1_1.net_value
        contrib_2 = a2_1.weighting * price2_2.net_value / price2_1.net_value
        contrib_3 = a3_1.weighting * price3_2.net_value / price3_1.net_value
        assert a1_2.weighting == pytest.approx(contrib_1 / (contrib_1 + contrib_2 + contrib_3), rel=Decimal(1e4))
        assert a2_2.weighting == pytest.approx(contrib_2 / (contrib_1 + contrib_2 + contrib_3), rel=Decimal(1e4))
        assert a3_2.weighting == pytest.approx(contrib_3 / (contrib_1 + contrib_2 + contrib_3), rel=Decimal(1e4))

        # Test if a deleted assets is kept if delete_existing_assets is set to True
        a1_1.delete()
        portfolio.propagate_or_update_assets(weekday, next_day, delete_existing_assets=True)
        with pytest.raises(AssetPosition.DoesNotExist):
            a1_2.refresh_from_db()

        a2_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i2)
        a3_2 = AssetPosition.objects.get(portfolio=portfolio, date=next_day, underlying_instrument=i3)
        assert a2_2
        assert a3_2

        # Test that we don't do anything on target portfolio because there is a non estimated position
        a2_2.is_estimated = False
        a2_2.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        a2_2_weighting = a2_2.weighting
        a3_2_weighting = a3_2.weighting
        a2_2.refresh_from_db()
        a3_2.refresh_from_db()

        assert a2_2.weighting == a2_2_weighting
        assert a3_2.weighting == a3_2_weighting

    def test_propagate_or_update_assets_active_states(
        self, weekday, active_product, asset_position_factory, instrument
    ):
        next_day = (weekday + BDay(1)).date()

        portfolio = active_product.portfolio

        a1 = asset_position_factory.create(
            portfolio=portfolio, date=weekday, underlying_instrument=instrument, currency=instrument.currency
        )
        asset_position_factory.create(
            portfolio=portfolio,
            date=next_day,
            underlying_instrument=instrument,
            currency=instrument.currency,
            exchange=a1.exchange,
            portfolio_created=a1.portfolio_created,
        )
        active_product.delisted_date = weekday
        active_product.save()
        # Test1: test if unactive portfolio keep having the to date assets. (asset found at next day are suppose to be deleted when the portfolio is non active at the from date)
        portfolio.propagate_or_update_assets(weekday, next_day)
        assert portfolio.assets.filter(date=next_day).exists() is False

        # Activate product
        active_product.delisted_date = None
        active_product.save()

        # Expect proper creation
        portfolio.propagate_or_update_assets(weekday, next_day)
        a_future = AssetPosition.objects.get(portfolio=portfolio, date=next_day)

        # Test if only estimated update existing pos,

        portfolio.propagate_or_update_assets(weekday, next_day)
        initial_shares = a1.initial_shares

        # Test that estimated shares keep being updated
        a1.initial_shares *= 2
        a1.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        a_future.refresh_from_db()
        assert a_future.initial_shares == initial_shares * 2

        # Test that non-estimated shares are not being updated
        updated_fields = ["initial_currency_fx_rate", "weighting", "initial_price", "initial_shares"]
        for field in updated_fields:
            setattr(a1, field, getattr(a1, field) * 2)
        a1.save()
        a_future_copy = model_to_dict(a_future)
        a_future.is_estimated = False
        a_future.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        a_future.refresh_from_db()
        for field in updated_fields:
            assert getattr(a_future, field) == a_future_copy[field]

        # Test active (from) portfolio but not active (to) create a zero weight position
        active_product.delisted_date = next_day
        active_product.save()

        a_future.is_estimated = True
        a_future.save()
        portfolio.propagate_or_update_assets(weekday, next_day)
        a_future.refresh_from_db()
        assert a_future.weighting == 0

    @patch.object(PortfolioSynchronization, "synchronize")
    @pytest.mark.parametrize("timedelta_days", [fake.pyint(min_value=1, max_value=10)])
    def test_resynchronize_history(
        self, mock_fct, portfolio, asset_position_factory, weekday, timedelta_days, portfolio_synchronization
    ):
        portfolio.portfolio_synchronization = portfolio_synchronization
        portfolio.save()
        asset_position_factory.create(portfolio=portfolio, date=weekday)
        portfolio.resynchronize_history(weekday, (weekday + BDay(timedelta_days)).date())
        mock_fct.call_count == timedelta_days
        with pytest.raises(ValueError):
            portfolio.resynchronize_history((weekday + BDay(timedelta_days)).date(), weekday)

    def test_update_preferred_classification_per_instrument(
        self, portfolio, asset_position_factory, equity_factory, classification_factory, classification_group_factory
    ):
        primary_group = classification_group_factory.create(is_primary=True)
        other_group = classification_group_factory.create(is_primary=False)
        c1 = classification_factory.create(group=other_group)
        c2_primary = classification_factory.create(group=primary_group)
        c2_secondary = classification_factory.create(group=other_group)
        c3_1 = classification_factory.create(group=other_group)
        c3_2 = classification_factory.create(group=other_group)

        # One classification to this instrument, we expect the relationship to be filled automatically
        i1 = equity_factory.create(classifications=[c1])
        # One classification "Primary" and one "other" to this instrument, we expect the relationship to be filled automatically
        i2 = equity_factory.create(classifications=[c2_secondary, c2_primary])
        # Two non-primary classifications to this instrument, we expect the relationship to not be filled with the classification automatically (created though)
        i3 = equity_factory.create(classifications=[c3_2, c3_1])
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=i1)
        a2 = asset_position_factory.create(portfolio=portfolio, underlying_instrument=i2)
        a3 = asset_position_factory.create(portfolio=portfolio, underlying_instrument=i3)

        assert not portfolio.preferred_instrument_classifications.exists()
        portfolio.update_preferred_classification_per_instrument()
        res1 = PortfolioInstrumentPreferredClassificationThroughModel.objects.get(
            portfolio=portfolio, classification=c1, instrument=i1, classification_group=other_group
        )
        res2 = PortfolioInstrumentPreferredClassificationThroughModel.objects.get(
            portfolio=portfolio, classification=c2_secondary, instrument=i2, classification_group=other_group
        )
        res3 = PortfolioInstrumentPreferredClassificationThroughModel.objects.get(
            portfolio=portfolio, classification=None, instrument=i3, classification_group=None
        )

        assert not PortfolioInstrumentPreferredClassificationThroughModel.objects.exclude(
            id__in=[res1.id, res2.id, res3.id]
        ).exists()

        # We delete portfolio positions and retrigger the function to check that the leftovers relationship are indeed removed
        a3.delete()
        a2.delete()
        portfolio.update_preferred_classification_per_instrument()
        with pytest.raises(PortfolioInstrumentPreferredClassificationThroughModel.DoesNotExist):
            res3.refresh_from_db()
        with pytest.raises(PortfolioInstrumentPreferredClassificationThroughModel.DoesNotExist):
            res2.refresh_from_db()
        res1.refresh_from_db()
        assert res1

    def test_import_positions_at_date(self, portfolio_factory, asset_position_factory, instrument_factory, weekday):
        def _serialize(obj):
            return {
                "portfolio": obj.portfolio.id,
                "portfolio_created": obj.portfolio_created.id if obj.portfolio_created else None,
                "underlying_instrument": obj.underlying_instrument.id,
                "date": obj.date,
                "currency": obj.currency.id,
                "weighting": obj.weighting,
                "initial_currency_fx_rate": obj.initial_currency_fx_rate,
                "initial_shares": obj.initial_shares,
                "initial_price": obj.initial_price,
                "is_estimated": obj.is_estimated,
                "exchange": obj.exchange.id if obj.exchange else None,
                "asset_valuation_date": obj.asset_valuation_date,
            }

        portfolio = portfolio_factory.create()

        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        a1 = asset_position_factory.build(
            portfolio=portfolio, date=weekday, underlying_instrument=i1, currency=i1.currency, weighting=0.5
        )
        a2 = asset_position_factory.build(
            portfolio=portfolio, date=weekday, underlying_instrument=i2, currency=i2.currency, weighting=0.25
        )
        a3 = asset_position_factory.build(
            portfolio=portfolio,
            currency=i2.currency,
            underlying_instrument=i2,
            initial_price=a2.initial_price,
            date=weekday,
            weighting=0.25,
        )

        portfolio.import_positions_at_date(
            PortfolioDTO([a1._build_dto(), a2._build_dto(), a3._build_dto()]),
            weekday,
        )

        res1 = AssetPosition.objects.get(portfolio=portfolio, date=weekday, underlying_instrument=i1)
        res2 = AssetPosition.objects.get(portfolio=portfolio, date=weekday, underlying_instrument=i2)

        assert portfolio.assets.filter(date=weekday).count() == 2

        assert res1.initial_shares == a1.initial_shares
        assert res1.weighting == a1.weighting / (a1.weighting + a2.weighting + a3.weighting)
        assert res1.initial_currency_fx_rate == a1.initial_currency_fx_rate
        assert res1.initial_price == a1.initial_price

        assert res2.initial_shares == a2.initial_shares + a3.initial_shares
        assert res2.weighting == a2.weighting + a3.weighting
        assert res2.initial_currency_fx_rate == (a2.initial_currency_fx_rate + a3.initial_currency_fx_rate) / 2
        assert res2.initial_price == (a2.initial_price + a3.initial_price) / 2

        assert portfolio.assets.filter(date=weekday, underlying_instrument=a2.underlying_instrument).count() == 1
        assert not portfolio.assets.filter(
            date=weekday, underlying_instrument=a2.underlying_instrument, initial_shares=a3.initial_shares
        ).exists()

        portfolio.import_positions_at_date(
            PortfolioDTO([a1._build_dto()]),
            weekday,
        )
        res1.refresh_from_db()
        with pytest.raises(AssetPosition.DoesNotExist):
            res2.refresh_from_db()

    def test_get_total_value(
        self, portfolio, customer_trade_factory, instrument_factory, instrument_price_factory, weekday
    ):
        i1 = instrument_factory.create()
        i2 = instrument_factory.create()
        previous_day = (weekday - BDay(5)).date()
        price11 = instrument_price_factory.create(instrument=i1, date=weekday, calculated=False)
        price12 = instrument_price_factory.create(instrument=i1, date=previous_day, calculated=False)
        price2 = instrument_price_factory.create(instrument=i2, date=weekday, calculated=False)

        # "noise" trades
        pending_customer_trade_i1 = customer_trade_factory.create(  # noqa
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i1, pending=True
        )
        marked_for_deletion_customer_trade_i1 = customer_trade_factory.create(  # noqa
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i1, marked_for_deletion=True
        )

        # valid trade for two different instrument but within the same portfolio
        trade_11 = customer_trade_factory.create(
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i1
        )
        trade_12 = customer_trade_factory.create(
            portfolio=portfolio, transaction_date=previous_day, underlying_instrument=i1
        )
        trade_2 = customer_trade_factory.create(
            portfolio=portfolio, transaction_date=weekday, underlying_instrument=i2
        )

        assert (
            portfolio.get_total_value(weekday)
            == price11.net_value * (trade_11.shares + trade_12.shares) + price2.net_value * trade_2.shares
        )
        assert portfolio.get_total_value(previous_day) == price12.net_value * trade_12.shares
        assert portfolio.get_total_value(previous_day - BDay(1)) == Decimal(0)

    def test_tracked_object(self, portfolio, asset_position_factory):
        assert not Portfolio.tracked_objects.exists()

        asset_position_factory.create(portfolio=portfolio)
        assert set(Portfolio.tracked_objects.all()) == {
            portfolio,
        }

        portfolio.is_tracked = False
        portfolio.save()
        assert not Portfolio.tracked_objects.exists()

    def test_is_invested_at_date(self, portfolio_factory):
        portfolio = portfolio_factory.create(invested_timespan=DateRange(date(2024, 1, 2), date(2024, 1, 3)))
        assert portfolio.is_invested_at_date(date(2024, 1, 1)) is False
        assert portfolio.is_invested_at_date(date(2024, 1, 2)) is True
        assert portfolio.is_invested_at_date(date(2024, 1, 3)) is False

        assert set(Portfolio.objects.filter_invested_at_date(date(2024, 1, 1))) == set()
        assert set(Portfolio.objects.filter_invested_at_date(date(2024, 1, 2))) == {portfolio}
        assert set(Portfolio.objects.filter_invested_at_date(date(2024, 1, 1))) == set()



from datetime import date
from decimal import Decimal
from typing import Any, Dict, Optional

import pandas as pd
from wbportfolio.models import Portfolio


def callback(
    portfolio: Portfolio,
    sync_date: date,
    rebalancing_freq: Optional[str] = "B",
    equally_weighted: Optional[bool] = False,
    composite: Optional[bool] = False,
    base_assets: Optional[Dict] = dict(),
    **kwargs: Any,
):
    """Recursively calculates the position for a portfolio

    Arguments:
        portfolio {portfolio.Portfolio} -- The Portfolio on which the assets will be computed
        sync_date {datetime.date} -- The date on which the assets will be computed

    Keyword Arguments:
        portfolio {portfolio.Portfolio} -- The core portfolio from which the computed position are created (default: {None})
        adjusted_weighting {int} -- the adjusted weight of the current level of index (default: {1})
        adjusted_currency_fx_rate {int} -- the adjusted currency exchange rate on the current level of index (default: {1})

    Yields:
        tuple[dict, dict] -- Two dictionaries: One with filter parameters and one with default values
    """
    assets = portfolio.assets.filter(date=sync_date)
    if composite:
        last_trade_proposals = portfolio.trade_proposals.filter(trade_date__lte=sync_date)
        if last_trade_proposals.exists():
            base_assets = last_trade_proposals.latest("trade_date").base_assets
    if assets.exists() and assets.filter(date=sync_date).exists():
        for asset in assets.all():
            new_weight = asset.weighting
            if pd.date_range(end=sync_date, periods=1, freq=rebalancing_freq)[0] == pd.Timestamp(sync_date):
                if equally_weighted:
                    new_weight = Decimal(1 / assets.count())
                elif base_assets and (proposed_weight := base_assets.get(asset.underlying_instrument.id, None)):
                    new_weight = proposed_weight
            yield asset._build_dto(), asset._build_dto(new_weight)



from contextlib import suppress
from datetime import timedelta
from typing import TypeVar

import pandas as pd
from celery import shared_task
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.functional import cached_property
from django_fsm import FSMField, transition
from pandas.tseries.offsets import BDay
from wbcompliance.models.risk_management.mixins import RiskCheckMixin
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models import WBModel
from wbfdm.models.instruments.instruments import Instrument
from wbportfolio.models.roles import PortfolioRole
from wbportfolio.pms.trading import TradingService
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import TradeBatch as TradeBatchDTO

from .trades import Trade

SelfTradeProposal = TypeVar("SelfTradeProposal", bound="TradeProposal")


class TradeProposal(RiskCheckMixin, WBModel):
    trade_date = models.DateField(verbose_name="Trading Date")

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        SUBMIT = "SUBMIT", "Submit"
        APPROVED = "APPROVED", "Approved"
        DENIED = "DENIED", "Denied"

    comment = models.TextField(default="", verbose_name="Trade Comment", blank=True)
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    model_portfolio = models.ForeignKey(
        "wbportfolio.Portfolio",
        blank=True,
        null=True,
        related_name="model_trade_proposals",
        on_delete=models.PROTECT,
        verbose_name="Model Portfolio",
    )
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", related_name="trade_proposals", on_delete=models.PROTECT, verbose_name="Portfolio"
    )
    creator = models.ForeignKey(
        "directory.Person",
        blank=True,
        null=True,
        related_name="trade_proposals",
        on_delete=models.PROTECT,
        verbose_name="Owner",
    )

    def _get_checked_object_field_name(self) -> str:
        """
        Mandatory function from the Riskcheck mixin that returns the field (aka portfolio), representing the object to check the rules against.
        """
        return "portfolio"

    @cached_property
    def validated_trading_service(self) -> TradingService:
        """
        This property holds the validated trading services and cache it.This property expect to be set only if is_valid return True
        """
        return TradingService(
            self.trade_date,
            effective_portfolio=self.portfolio._build_dto(self.trade_date),
            trades_batch=self._build_dto(),
        )

    @property
    def previous_trade_proposal(self) -> SelfTradeProposal | None:
        future_proposals = TradeProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__lt=self.trade_date, status=TradeProposal.Status.APPROVED
        )
        if future_proposals.exists():
            return future_proposals.latest("trade_date")
        return None

    @property
    def next_trade_proposal(self) -> SelfTradeProposal | None:
        future_proposals = TradeProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__gt=self.trade_date, status=TradeProposal.Status.APPROVED
        )
        if future_proposals.exists():
            return future_proposals.earliest("trade_date")
        return None

    @cached_property
    def base_assets(self):
        """
        Return a dictionary representation (instrument_id: target weight) of this trade proposal
        Returns:
            A dictionary representation

        """
        return {
            v["underlying_instrument"]: v["target_weight"]
            for v in self.trades.filter(status=Trade.Status.EXECUTED).values("underlying_instrument", "target_weight")
        }

    def __str__(self) -> str:
        return f"{self.portfolio.name}: {self.trade_date} ({self.status})"

    def save(self, *args, **kwargs):
        if not self.model_portfolio:
            self.model_portfolio = self.portfolio
        super().save(*args, **kwargs)

    def _build_dto(self) -> TradeBatchDTO:
        """
        Data Transfer Object
        Returns:
            DTO trade object
        """
        return (
            TradeBatchDTO(tuple([trade._build_dto() for trade in self.trades.all()])) if self.trades.exists() else None
        )

    # Start tools methods
    def clone(self, **kwargs) -> SelfTradeProposal:
        """
        Method to clone self as a new trade proposal. It will automatically shift the trade date if a proposal already exists
        Args:
            **kwargs: The keyword arguments
        Returns:
            The cloned trade proposal
        """
        trade_date = kwargs.get("trade_date", self.trade_date)

        # Find the next valid trade date
        while TradeProposal.objects.filter(portfolio=self.portfolio, trade_date=trade_date).exists():
            trade_date += timedelta(days=1)

        trade_proposal_clone = TradeProposal.objects.create(
            trade_date=trade_date,
            comment=kwargs.get("comment", self.comment),
            status=TradeProposal.Status.DRAFT,
            model_portfolio=self.model_portfolio,
            portfolio=self.portfolio,
            creator=self.creator,
        )

        # For all existing trades, copy them to the new trade proposal
        for trade in self.trades.all():
            trade.pk = None
            trade.trade_proposal = trade_proposal_clone
            trade.transaction_date = trade_proposal_clone.trade_date
            trade.save()
        return trade_proposal_clone

    def normalize_trades(self):
        """
        Call the trading service with the existing trades and normalize them in order to obtain a total sum target weight of 100%
        The existing trade will be modified directly with the given normalization factor
        """
        service = TradingService(self.trade_date, trades_batch=self._build_dto())
        service.normalize()
        leftovers_trades = self.trades.all()
        for _, trade in service.trades_batch.trades_map.items():
            with suppress(Trade.DoesNotExist):
                self.trades.update_or_create(
                    id=trade.id,
                    defaults={
                        "weighting": trade.delta_weight,
                        "shares": trade.target_shares,
                    },
                )
                leftovers_trades = leftovers_trades.exclude(id=trade.id)
        leftovers_trades.delete()

    def reset_trades(self):
        """
        Will delete all existing trades and recreate them from the method `create_or_update_trades`
        """
        # delete all existing trades
        self.trades.all().delete()
        # recreate them from scratch (if the portfolio has positions)
        self.create_or_update_trades()

    def apply_trades(self):
        # We validate trade which will create or update the initial asset positions
        self.trades.exclude(status=Trade.Status.SUBMIT).update(status=Trade.Status.SUBMIT)
        for trade in self.trades.all():
            trade.execute()
            trade.save()
        # We propagate the new portfolio composition until the next trade proposal or today if it doesn't exist yet
        to_date = self.next_trade_proposal.trade_date if self.next_trade_proposal else timezone.now().date()

        for from_date in pd.date_range(self.trade_date, to_date - timedelta(days=1), freq="B"):
            to_date = (from_date + BDay(1)).date()
            self.portfolio.propagate_or_update_assets(
                from_date.date(),
                to_date,
                forward_price=False,
                base_assets=self.base_assets,
                delete_existing_assets=True,
            )
            self.portfolio.change_at_date(to_date, base_assets=self.base_assets, force_recompute_weighting=True)

    def revert_trades(self):
        self.trades.exclude(status=Trade.Status.EXECUTED).update(status=Trade.Status.EXECUTED)
        for trade in self.trades.all():
            trade.revert()
            trade.save()
        if previous_trade_proposal := self.previous_trade_proposal:
            previous_trade_proposal.apply_trades()

    def create_or_update_trades(
        self, target_portfolio: PortfolioDTO = None, effective_portfolio: PortfolioDTO = None, reset: bool = False
    ):
        """
        This function talk to the trading service layer in order to generate a list of valid trades to attach to the proposal

        Args:
            target_portfolio: The target portfolio that the trades needs to execute to. Absence of position means a sell
            effective_portfolio: The current or effective portfolio to derivative effective weight from. Absence of position means a buy
            reset: If true, delete the current attached trades
        """
        # if the target portfolio is not provided, we try to build it
        if (
            not target_portfolio
            and (assets := self.model_portfolio.assets.filter(date__lte=self.trade_date)).exists()
            and (latest_pos := assets.latest("date"))
        ):
            target_portfolio = self.model_portfolio._build_dto(latest_pos.date)

        # if the effective portfolio is not provided, we try to build it
        if (
            not effective_portfolio
            and (assets := self.portfolio.assets.filter(date__lte=self.trade_date)).exists()
            and (latest_pos := assets.latest("date"))
        ):
            effective_portfolio = self.portfolio._build_dto(latest_pos.date)
        # Build trades DTO from the attached trades
        trade_batch = self._build_dto()
        if target_portfolio or effective_portfolio or trade_batch:
            service = TradingService(
                self.trade_date,
                effective_portfolio=effective_portfolio,
                target_portfolio=target_portfolio,
                trades_batch=trade_batch,
            )
            # with suppress(ValidationError):
            # Normalize the trades and validate it
            service.normalize()
            service.is_valid()
            if reset:
                self.trades.all().delete()
            for trade_dto in service.validated_trades:
                instrument = Instrument.objects.get(id=trade_dto.underlying_instrument)
                t, c = Trade.objects.update_or_create(
                    underlying_instrument=instrument,
                    currency=instrument.currency,
                    transaction_date=self.trade_date,
                    trade_proposal=self,
                    portfolio=self.portfolio,
                    defaults={
                        "shares": trade_dto.target_shares,
                        "weighting": trade_dto.delta_weight,
                        "status": Trade.Status.DRAFT,
                        "currency_fx_rate": instrument.currency.convert(self.trade_date, self.portfolio.currency),
                    },
                )

    # End tools methods

    # Start FSM logics

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.SUBMIT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submit",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def submit(self, by=None, description=None, **kwargs):
        self.trades.update(status=Trade.Status.SUBMIT)
        self.evaluate_active_rules(
            self.trade_date, self.validated_trading_service.target_portfolio, asynchronously=True
        )

    def can_submit(self):
        errors = dict()
        errors_list = []
        if self.trades.exists() and self.trades.exclude(status=Trade.Status.DRAFT).exists():
            errors_list.append("All trades need to be draft before submitting")
        service = self.validated_trading_service
        try:
            service.is_valid(ignore_error=True)
            # if service.trades_batch.totat_abs_delta_weight == 0:
            #     errors_list.append(
            #         "There is no change detected in this trade proposal. Please submit at last one valid trade"
            #     )
            if len(service.validated_trades) == 0:
                errors_list.append("There is no valid trade on this proposal")
            if service.errors:
                errors_list.extend(service.errors)
            if errors_list:
                errors["non_field_errors"] = errors_list
        except ValidationError:
            errors["non_field_errors"] = service.errors
            with suppress(KeyError):
                del self.__dict__["validated_trading_service"]
        return errors

    @property
    def can_be_approved_or_denied(self):
        return self.has_no_rule_or_all_checked_succeed and self.portfolio.is_manageable

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.APPROVED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.can_be_approved_or_denied,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approve",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def approve(self, by=None, description=None, **kwargs):
        apply_trades_proposal_as_task.delay(self.id)

    def can_approve(self):
        errors = dict()
        if self.trades.exclude(status=Trade.Status.SUBMIT).exists():
            errors["non_field_errors"] = "At least one trade needs to be submitted to be able to approve this proposal"
        if not self.portfolio.is_manageable:
            errors[
                "portfolio"
            ] = "The portfolio needs to be a model portfolio in order to approve this trade proposal manually"
        if self.has_assigned_active_rules and not self.has_all_check_completed_and_succeed:
            errors["non_field_errors"] = "The pre trades rules did not passed successfully"
        return errors

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.DENIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.can_be_approved_or_denied,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        self.trades.all().delete()
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]

    def can_deny(self):
        errors = dict()
        if self.trades.exclude(status=Trade.Status.SUBMIT).exists():
            errors["non_field_errors"] = "At least one trade needs to be submitted to be able to deny this proposal"
        return errors

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.has_all_check_completed,  # we wait for all checks to succeed before proposing the back to draft transition
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="backtodraft",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        self.trades.update(status=Trade.Status.DRAFT)
        self.checks.delete()

    def can_backtodraft(self):
        errors = dict()
        if self.trades.exclude(status=Trade.Status.SUBMIT).exists():
            errors["non_field_errors"] = "All trades need to be submitted before reverting back to draft"
        return errors

    @transition(
        field=status,
        source=Status.APPROVED,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:tradeproposal",),
                icon=WBIcon.REGENERATE.icon,
                key="revert",
                label="Revert",
                action_label="revert",
                description_fields="<p>Unapply trades and move everything back to draft (i.e. The underlying asset positions will change like the trades were never applied)</p>",
            )
        },
    )
    def revert(self, **kwargs):
        with suppress(KeyError):
            del self.__dict__["validated_trading_service"]
        revert_trade_proposal_as_task.delay(self.id, **kwargs)

    def can_revert(self):
        errors = dict()
        if self.trades.exclude(status=Trade.Status.EXECUTED).exists():
            errors["non_field_errors"] = "All trades need to be executed before reverting"
        if not self.portfolio.is_manageable:
            errors[
                "portfolio"
            ] = "The portfolio needs to be a model portfolio in order to revert this trade proposal manually"
        return errors

    # End FSM logics

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbportfolio:tradeproposal"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:tradeproposalrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{_portfolio.name}} ({{trade_date}})"

    class Meta:
        verbose_name = "Trade Proposal"
        verbose_name_plural = "Trade Proposals"
        unique_together = ["portfolio", "trade_date"]


@shared_task(queue="portfolio")
def apply_trades_proposal_as_task(trade_proposal_id):
    trade_proposal = TradeProposal.objects.get(id=trade_proposal_id)
    trade_proposal.apply_trades()


@shared_task(queue="portfolio")
def revert_trade_proposal_as_task(trade_proposal_id, **kwargs):
    trade_proposal = TradeProposal.objects.get(id=trade_proposal_id)
    trade_proposal.revert_trades()


@receiver(post_save, sender="wbportfolio.TradeProposal")
def post_save_trade_proposal(sender, instance, created, raw, **kwargs):
    if created and not raw and instance.portfolio.assets.filter(date__lte=instance.trade_date).exists():
        instance.create_or_update_trades(reset=True)



from wbcore import serializers as wb_serializers
from wbportfolio.models import PortfolioSynchronization, PriceComputation


class PortfolioSynchronizationRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:portfolio-detail")

    class Meta:
        model = PortfolioSynchronization
        fields = ("id", "name", "import_path", "_detail")


class PriceComputationRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:portfolio-detail")

    class Meta:
        model = PriceComputation
        fields = ("id", "name", "import_path", "_detail")



from django_celery_beat.admin import (
    PeriodicTaskAdmin,
    PeriodicTaskForm,
    _,
    loads,
    messages,
    pluralize,
)


class SynchronizationTaskForm(PeriodicTaskForm):
    def clean(self):
        return self.cleaned_data


class SynchronizationAdmin(PeriodicTaskAdmin):
    form = SynchronizationTaskForm
    search_fields = ("name",)
    regtask = None
    readonly_fields = (
        "last_run_at",
        # 'regtask',
        "task",
        "args",
        # 'kwargs',
        "expires",
        "expire_seconds",
        "queue",
        "exchange",
        "routing_key",
        "priority",
        "headers",
    )
    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "task",
                    "enabled",
                    "description",
                ),
                "classes": ("extrapretty", "wide"),
            },
        ),
        (
            "Schedule",
            {
                "fields": ("crontab", "start_time", "last_run_at", "one_off"),
                "classes": ("extrapretty", "wide"),
            },
        ),
        (
            "Arguments",
            {
                "fields": ("args", "kwargs"),
                "classes": ("extrapretty", "wide", "collapse", "in"),
            },
        ),
        (
            "Execution Options",
            {
                "fields": ("expires", "expire_seconds", "queue", "exchange", "routing_key", "priority", "headers"),
                "classes": ("extrapretty", "wide", "collapse", "in"),
            },
        ),
    )

    def run_tasks(self, request, queryset):
        def _load_kwargs(task):
            kwargs = loads(task.kwargs)
            kwargs["override_execution_datetime_validity"] = True
            return kwargs

        self.celery_app.loader.import_default_modules()

        tasks = [
            (self.celery_app.tasks.get(task.task), loads(task.args), _load_kwargs(task), task.queue)
            for task in queryset
        ]

        if any(t[0] is None for t in tasks):
            for i, t in enumerate(tasks):
                if t[0] is None:
                    break

            # variable "i" will be set because list "tasks" is not empty
            not_found_task_name = queryset[i].task

            self.message_user(
                request,
                _('task "{0}" not found'.format(not_found_task_name)),
                level=messages.ERROR,
            )
            return

        task_ids = [
            task.apply_async(args=args, kwargs=kwargs, queue=queue)
            if queue and len(queue)
            else task.apply_async(args=args, kwargs=kwargs)
            for task, args, kwargs, queue in tasks
        ]
        tasks_run = len(task_ids)
        self.message_user(
            request,
            _("{0} task{1} {2} successfully run").format(
                tasks_run,
                pluralize(tasks_run),
                pluralize(tasks_run, _("was,were")),
            ),
        )

    run_tasks.short_description = _("Run selected tasks")



from datetime import date

from django.shortcuts import get_object_or_404
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbportfolio.models import Portfolio, PriceComputation
from wbportfolio.serializers import (
    PortfolioSynchronizationRepresentationSerializer,
    PriceComputationRepresentationSerializer,
)


def get_portfolio_start_end_serializer_class(portfolio):
    today = date.today()

    class StartEndDateSerializer(wb_serializers.Serializer):
        start = wb_serializers.DateField(
            label="Start",
            default=portfolio.assets.earliest("date").date if portfolio.assets.exists() else today,
        )
        end = wb_serializers.DateField(
            label="End", default=portfolio.assets.latest("date").date if portfolio.assets.exists() else today
        )

    return StartEndDateSerializer


class PortfolioButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        portfolio_buttons = [
            bt.WidgetButton(key="assets", label="Assets"),
            bt.WidgetButton(key="contributor", label="Contributor"),
            bt.WidgetButton(key="distribution_chart", label="Distribution Chart"),
            bt.WidgetButton(key="distribution_table", label="Distribution Table"),
            bt.WidgetButton(key="modelcomposition", label="Portfolio Composition vs. Dependant Portfolios"),
        ]
        if pk := self.view.kwargs.get("pk", None):
            portfolio = get_object_or_404(Portfolio, pk=pk)

            portfolio_buttons.append(
                bt.ActionButton(
                    method=RequestType.PATCH,
                    identifiers=("wbportfolio:portfolio",),
                    key="resynchronize",
                    label="Resynchronize/Rebalance",
                    description_fields="""
                    <p>Resynchronize portfolio from {{start}} to {{end}}</p>
                    """,
                    serializer=get_portfolio_start_end_serializer_class(portfolio),
                    action_label="resynchronize",
                    title="Resynchronize/Rebalance",
                    instance_display=create_simple_display([["start"], ["end"]]),
                )
            )

        return {
            bt.DropDownButton(label="Portfolio", icon=WBIcon.UNFOLD.icon, buttons=tuple(portfolio_buttons)),
        }

    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()


class CreateModelPortfolioSerializer(wb_serializers.ModelSerializer):
    create_index = wb_serializers.BooleanField(default=False, label="Create Underlying Index")
    name = wb_serializers.CharField(required=True)
    _currency = CurrencyRepresentationSerializer(source="currency")
    _portfolio_synchronization = PortfolioSynchronizationRepresentationSerializer(source="portfolio_synchronization")
    price_computation = wb_serializers.PrimaryKeyRelatedField(queryset=PriceComputation.objects.all(), required=False)
    _price_computation = PriceComputationRepresentationSerializer(source="price_computation")

    class Meta:
        model = Portfolio
        dependency_map = {
            "price_computation": ["create_index"],
        }
        fields = (
            "name",
            "currency",
            "portfolio_synchronization",
            "create_index",
            "price_computation",
            "_price_computation",
            "_currency",
            "_portfolio_synchronization",
        )


class ModelPortfolioButtonConfig(PortfolioButtonConfig):
    # pass
    def get_custom_buttons(self):
        if not self.view.kwargs.get("pk", None):
            return {
                bt.ActionButton(
                    method=RequestType.POST,
                    identifiers=("wbportfolio:portfolio",),
                    endpoint=reverse("wbportfolio:modelportfolio-createmodelportfolio", request=self.request),
                    label="Create New Model Portfolio",
                    serializer=CreateModelPortfolioSerializer,
                    action_label="create",
                    title="Create Model Portfolio",
                    instance_display=create_simple_display(
                        [["name", "currency", "portfolio_synchronization"], ["create_index", "price_computation", "."]]
                    ),
                )
            }
        return set()



from datetime import date, timedelta
from decimal import Decimal
from unittest.mock import PropertyMock, patch

import pytest
from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from django.db.models import Sum
from faker import Faker
from pandas.tseries.offsets import BDay
from wbcore.contrib.currency.models import Currency
from wbportfolio.models import Trade
from wbportfolio.models.synchronization.portfolio_synchronization import (
    PortfolioSynchronization,
    synchronize_portfolio_as_task,
)
from wbportfolio.models.synchronization.price_computation import (
    PriceComputation,
    compute_price_as_task,
)
from wbportfolio.pms.typing import Position as PositionDTO

from .utils import PortfolioTestMixin

fake = Faker()


@shared_task()
def callback(*args, **kwargs):
    return True


@pytest.mark.django_db
class TestSynchronizationTaskModel(PortfolioTestMixin):
    def test_init(self, synchronization_task):
        assert synchronization_task.id is not None

    def test_args(self, synchronization_task):
        content_type = ContentType.objects.get_for_model(synchronization_task)
        assert (
            synchronization_task.args
            == f'["{synchronization_task.id}", "{content_type.app_label}", "{content_type.model}"]'
        )

    def test_import_method(self, synchronization_task):
        with pytest.raises(NotImplementedError):
            method = synchronization_task._import_method
            method()

    def test_import_method_with_custom_callback(self, synchronization_task_factory):
        synchronization_task = synchronization_task_factory.create(
            import_path="wbportfolio.tests.models.test_synchronization"
        )
        callback_si_task = callback.si().task.replace("modules.wbportfolio.", "")
        test_callback_si_task = synchronization_task._import_method.si().task.replace("modules.wbportfolio.", "")
        assert callback_si_task == test_callback_si_task

    def test_tasks_signature(self, synchronization_task):
        with pytest.raises(NotImplementedError):
            synchronization_task._tasks_signature()

    def test_default_callback(self, synchronization_task):
        with pytest.raises(NotImplementedError):
            synchronization_task._default_callback()

    def test_end_task_signature(self, synchronization_task):
        assert synchronization_task._end_task_signature() is None

    def test_end_task_signature_with_signature(self, synchronization_task_factory, periodic_task_factory):
        dependent_task = synchronization_task_factory.create(
            task="wbportfolio.tests.models.test_synchronization.callback"
        )

        synchronization_task = synchronization_task_factory.create()
        synchronization_task.dependent_task = dependent_task
        synchronization_task.save()

        assert (
            synchronization_task._end_task_signature().task
            == "wbportfolio.models.synchronization.synchronization.task"
        )


@pytest.mark.django_db
class TestPortfolioSynchronization(PortfolioTestMixin):
    def test_init(self, portfolio_synchronization):
        assert portfolio_synchronization.id is not None

    @patch.object(PortfolioSynchronization, "_import_method", new_callable=PropertyMock)
    @patch.object(PortfolioSynchronization, "is_valid_date")
    def test_synchronize_automatic_approval(
        self,
        mock_is_valid_date,
        mock_import_method,
        portfolio_synchronization,
        portfolio_factory,
        asset_position_factory,
        equity_factory,
        currency_factory,
        trade_factory,
        instrument_portfolio_through_model_factory,
        active_product,
        weekday,
    ):
        portfolio = portfolio_factory.create(portfolio_synchronization=portfolio_synchronization)
        instrument_portfolio_through_model_factory.create(instrument=active_product, portfolio=portfolio)
        trade_factory.create(
            underlying_instrument=active_product,
            transaction_date=weekday,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            shares=100,
        )

        portfolio.save()

        def callback(*args, **kwargs):
            for i in range(10):
                position = asset_position_factory.build(
                    underlying_instrument=equity_factory.create(), currency=currency_factory.create()
                )
                yield None, PositionDTO(
                    underlying_instrument=position.underlying_instrument.id,
                    instrument_type=position.underlying_instrument.instrument_type,
                    date=weekday,
                    asset_valuation_date=weekday,
                    currency=position.currency.id,
                    price=position.initial_price,
                    shares=position.initial_shares,
                    currency_fx_rate=position.initial_currency_fx_rate,
                    weighting=position.weighting,
                    is_estimated=position.is_estimated,
                )

        mock_import_method.return_value = callback
        mock_is_valid_date.return_value = True
        portfolio_synchronization.synchronize(portfolio, weekday, override_execution_datetime_validity=True)
        assert portfolio.assets.count() == 10
        assert 1.0 == pytest.approx(float(portfolio.assets.aggregate(s=Sum("weighting"))["s"]), rel=1e-4)

    @patch.object(PortfolioSynchronization, "_import_method", new_callable=PropertyMock)
    @patch.object(PortfolioSynchronization, "is_valid_date")
    @patch.object(Currency, "convert")
    def test_synchronize_tradeprosal(
        self,
        mock_convert,
        mock_is_valid_date,
        mock_import_method,
        portfolio_synchronization_factory,
        portfolio_factory,
        asset_position_factory,
        equity_factory,
        currency_factory,
        active_product,
        weekday,
        trade_factory,
        instrument_portfolio_through_model_factory,
    ):
        portfolio_synchronization = portfolio_synchronization_factory(is_automatic_validation=False)

        portfolio = portfolio_factory.create()
        portfolio.portfolio_synchronization = portfolio_synchronization
        portfolio.save()

        instrument_portfolio_through_model_factory.create(instrument=active_product, portfolio=portfolio)
        trade_factory.create(
            underlying_instrument=active_product,
            transaction_date=weekday,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            shares=100,
        )

        def callback(*args, **kwargs):
            for i in range(10):
                position = asset_position_factory.build(
                    underlying_instrument=equity_factory.create(), currency=currency_factory.create()
                )
                position_dict = PositionDTO(
                    underlying_instrument=position.underlying_instrument.id,
                    instrument_type=position.underlying_instrument.instrument_type,
                    date=weekday,
                    asset_valuation_date=weekday,
                    currency=position.currency.id,
                    price=position.initial_price,
                    shares=position.initial_shares,
                    currency_fx_rate=position.initial_currency_fx_rate,
                    weighting=position.weighting,
                    is_estimated=position.is_estimated,
                )
                yield position_dict, position_dict

        mock_is_valid_date.return_value = True
        mock_import_method.return_value = callback
        mock_convert.return_value = Decimal(1.0)
        portfolio_synchronization.synchronize(portfolio, weekday, override_execution_datetime_validity=True)
        assert portfolio.trade_proposals.count() == 1
        assert portfolio.transactions.count() == 10

    def test_synchronize_as_task_si(self, portfolio_synchronization, portfolio, weekday):
        assert portfolio_synchronization.synchronize_as_task_si(
            portfolio, weekday
        ) == synchronize_portfolio_as_task.si(portfolio_synchronization.id, portfolio.id, weekday)

    def test_tasks_signature(
        self,
        portfolio_synchronization,
        portfolio_factory,
        weekday,
        trade_factory,
        product_factory,
        instrument_portfolio_through_model_factory,
    ):
        p1 = portfolio_factory.create()
        p1.portfolio_synchronization = portfolio_synchronization
        p1.save()
        product1 = product_factory.create(delisted_date=None, inception_date=weekday - timedelta(days=30))
        instrument_portfolio_through_model_factory.create(instrument=product1, portfolio=p1)
        trade_factory.create(
            underlying_instrument=product1,
            transaction_date=weekday,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            shares=100,
        )

        p2 = portfolio_factory.create()
        p2.portfolio_synchronization = portfolio_synchronization
        p2.save()
        product2 = product_factory.create(delisted_date=None, inception_date=weekday - timedelta(days=30))
        instrument_portfolio_through_model_factory.create(instrument=product2, portfolio=p2)
        trade_factory.create(
            underlying_instrument=product2,
            transaction_date=weekday,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            shares=100,
        )

        signatures = portfolio_synchronization._tasks_signature(sync_date=weekday)
        assert len(list(signatures)) == 2

    @patch.object(PortfolioSynchronization, "is_valid_date")
    def test_default_callback(
        self,
        mock_is_valid_date,
        active_product,
        weekday,
        portfolio_synchronization,
        portfolio_factory,
        index_factory,
        equity_factory,
        asset_position_factory,
        trade_factory,
        instrument_price_factory,
        instrument_portfolio_through_model_factory,
    ):
        while weekday.weekday() in [5, 6]:
            weekday += timedelta(days=1)
        mock_is_valid_date.return_value = True
        root_index = index_factory.create()
        root_index_portfolio = root_index.portfolio

        index1 = index_factory.create()
        index1_portfolio = index1.portfolio

        a1 = asset_position_factory.create(
            underlying_instrument=index1,
            portfolio=root_index_portfolio,
            weighting=0.6,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        index2 = index_factory.create()
        index2_portfolio = index2.portfolio

        a2 = asset_position_factory.create(
            underlying_instrument=index2,
            portfolio=root_index_portfolio,
            weighting=0.4,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        a1_1 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index1_portfolio,
            weighting=0.2,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )
        a2_1 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index1_portfolio,
            weighting=0.3,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )
        a3_1 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index1_portfolio,
            weighting=0.5,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        a1_2 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index2_portfolio,
            weighting=0.7,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )
        a2_2 = asset_position_factory.create(
            underlying_instrument=equity_factory.create(),
            portfolio=index2_portfolio,
            weighting=0.3,
            initial_shares=None,
            initial_price=100,
            date=weekday,
        )

        product_base_portfolio = active_product.primary_portfolio
        product_portfolio = portfolio_factory.create(portfolio_synchronization=portfolio_synchronization)
        instrument_portfolio_through_model_factory.create(instrument=active_product, portfolio=product_portfolio)
        trade_factory.create(
            underlying_instrument=active_product,
            transaction_date=weekday,
            transaction_subtype=Trade.Type.SUBSCRIPTION,
            shares=100,
        )

        product_portfolio.depends_on.add(root_index_portfolio)

        instrument_portfolio_through_model_factory.create(instrument=active_product, portfolio=product_portfolio)

        instrument_price_factory.create(instrument=active_product, date=weekday)
        trade_factory.create(
            underlying_instrument=active_product,
            portfolio=product_base_portfolio,
            transaction_date=weekday,
            shares=1000,
            transaction_subtype="SUBSCRIPTION",
        )

        portfolio_synchronization.synchronize(product_portfolio, weekday, override_execution_datetime_validity=True)
        assert product_portfolio.assets.filter(date=weekday).count() == 5
        assert float(a1_1.weighting) * float(a1.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index1_portfolio,
                    underlying_instrument=a1_1.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a2_1.weighting) * float(a1.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index1_portfolio,
                    underlying_instrument=a2_1.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a3_1.weighting) * float(a1.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index1_portfolio,
                    underlying_instrument=a3_1.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a1_2.weighting) * float(a2.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index2_portfolio,
                    underlying_instrument=a1_2.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert float(a2_2.weighting) * float(a2.weighting) == pytest.approx(
            float(
                product_portfolio.assets.filter(
                    portfolio_created=index2_portfolio,
                    underlying_instrument=a2_2.underlying_instrument,
                    date=weekday,
                )
                .first()
                .weighting
            )
        )
        assert Decimal(1.0) == pytest.approx(product_portfolio.assets.aggregate(s=Sum("weighting"))["s"])

    @patch.object(PortfolioSynchronization, "synchronize")
    def test_synchronize_portfolio_as_task(self, mock_synchronize, portfolio_synchronization, portfolio, weekday):
        synchronize_portfolio_as_task(portfolio_synchronization.id, portfolio.id, weekday)
        assert mock_synchronize.call_count == 1


@pytest.mark.django_db
class TestPriceComputation(PortfolioTestMixin):
    def test_init(self, price_computation):
        assert price_computation.id is not None

    def test_compute_price_as_task_si(self, price_computation, instrument, weekday):
        assert price_computation.compute_price_as_task_si(instrument, weekday) == compute_price_as_task.si(
            price_computation.id, instrument.id, weekday
        )

    def test_tasks_signature_with_sync_date(self, price_computation, product_factory, weekday):
        i1 = product_factory.create(inception_date=weekday - BDay(1), delisted_date=None)
        i1.price_computation = price_computation
        i1.save()

        i2 = product_factory.create(inception_date=weekday - BDay(1), delisted_date=None)
        i2.price_computation = price_computation
        i2.save()

        i_inactive = product_factory.create(
            inception_date=weekday - BDay(1), delisted_date=weekday - timedelta(days=1)
        )
        i_inactive.price_computation = price_computation
        i_inactive.save()

        signatures = price_computation._tasks_signature(sync_date=weekday)

        assert set(map(lambda x: (x.args[0], x.args[1], x.args[2]), signatures)) == {
            (price_computation.id, i1.id, weekday),
            (price_computation.id, i2.id, weekday),
        }

    def test_tasks_signature_without_sync_date(self, price_computation, product_factory, instrument_price_factory):
        thursday = date(2023, 10, 19)
        friday = date(2023, 10, 20)
        saturday = date(2023, 10, 21)
        sunday = date(2023, 10, 22)
        monday = date(2023, 10, 23)
        tuesday = date(2023, 10, 24)
        product = product_factory.create(inception_date=thursday, delisted_date=None)
        product.price_computation = price_computation
        product.save()
        instrument_price_factory.create(date=thursday, instrument=product)

        signatures_friday = price_computation._tasks_signature(to_date=friday)
        assert set(map(lambda x: (x.args[0], x.args[1], x.args[2]), signatures_friday)) == {
            (price_computation.id, product.id, thursday),
        }

        signatures_saturday = price_computation._tasks_signature(to_date=saturday)
        assert set(map(lambda x: (x.args[0], x.args[1], x.args[2]), signatures_saturday)) == {
            (price_computation.id, product.id, thursday),
            (price_computation.id, product.id, friday),
        }

        signatures_sunday = price_computation._tasks_signature(to_date=sunday)
        assert set(map(lambda x: (x.args[0], x.args[1], x.args[2]), signatures_sunday)) == {
            (price_computation.id, product.id, thursday),
            (price_computation.id, product.id, friday),
        }

        signatures_monday = price_computation._tasks_signature(to_date=monday)
        assert set(map(lambda x: (x.args[0], x.args[1], x.args[2]), signatures_monday)) == {
            (price_computation.id, product.id, thursday),
            (price_computation.id, product.id, friday),
        }

        signatures_tuesday = price_computation._tasks_signature(to_date=tuesday)
        assert set(map(lambda x: (x.args[0], x.args[1], x.args[2]), signatures_tuesday)) == {
            (price_computation.id, product.id, thursday),
            (price_computation.id, product.id, friday),
            (price_computation.id, product.id, monday),
        }
        instrument_price_factory.create(date=monday, instrument=product)
        signatures = price_computation._tasks_signature(to_date=tuesday)
        assert set(map(lambda x: (x.args[0], x.args[1], x.args[2]), signatures)) == {
            (price_computation.id, product.id, monday),
        }

    @patch.object(PriceComputation, "compute")
    def test_compute_price_as_task(self, mock_compute, price_computation, instrument, weekday):
        compute_price_as_task(price_computation.id, instrument.id, weekday)
        assert mock_compute.call_count == 1

    @patch.object(PriceComputation, "_import_method", new_callable=PropertyMock)
    def test_compute(
        self, mock_import_method, weekday, price_computation, instrument_factory, instrument_price_factory
    ):
        while weekday.weekday() in [5, 6]:
            weekday += timedelta(days=1)
        instrument = instrument_factory.create(inception_date=weekday - timedelta(days=1), delisted_date=None)
        instrument.price_computation = price_computation
        instrument.save()
        price = instrument_price_factory.build(instrument=instrument, date=weekday)

        def callback(*args, **kwargs):
            return {
                "instrument": instrument,
                "date": weekday,
                "gross_value": price.gross_value,
                "net_value": price.net_value,
            }

        mock_import_method.return_value = callback
        price_computation.compute(instrument, weekday, override_execution_datetime_validity=True)
        assert instrument.prices.count() == 1
        assert float(price.net_value) == pytest.approx(float(instrument.prices.first().net_value))

    def test_default_callback(
        self,
        weekday,
        price_computation,
        equity_factory,
        product_factory,
        asset_position_factory,
        instrument_price_factory,
        trade_factory,
    ):
        while weekday.weekday() in [5, 6]:
            weekday += timedelta(days=1)

        previous_sync_date = weekday - timedelta(days=1)
        while previous_sync_date.weekday() in [5, 6]:
            previous_sync_date -= timedelta(days=1)

        product = product_factory.create(
            price_computation=price_computation, inception_date=weekday - timedelta(days=1), delisted_date=None
        )
        portfolio = product.portfolio

        trade_factory.create(
            underlying_instrument=product,
            portfolio=portfolio,
            transaction_date=previous_sync_date,
            shares=1000,
            transaction_subtype="SUBSCRIPTION",
        )

        e1 = equity_factory.create()
        e2 = equity_factory.create()
        e3 = equity_factory.create()

        a1_1 = asset_position_factory.create(
            underlying_instrument=e1,
            portfolio=portfolio,
            date=previous_sync_date,
            weighting=Decimal(0.3),
            initial_price=Decimal(100),
            initial_shares=300,
        )
        a2_1 = asset_position_factory.create(
            underlying_instrument=e2,
            portfolio=portfolio,
            date=previous_sync_date,
            weighting=Decimal(0.5),
            initial_price=Decimal(100),
            initial_shares=500,
        )
        a3_1 = asset_position_factory.create(
            underlying_instrument=e3,
            portfolio=portfolio,
            date=previous_sync_date,
            weighting=Decimal(0.2),
            initial_price=Decimal(100),
            initial_shares=200,
        )

        a1_2 = asset_position_factory.create(
            underlying_instrument=e1,
            portfolio=portfolio,
            date=weekday,
            weighting=Decimal(0.3719),
            initial_price=Decimal(150),
            initial_shares=300,
        )
        a2_2 = asset_position_factory.create(
            underlying_instrument=e2,
            portfolio=portfolio,
            date=weekday,
            weighting=Decimal(0.4959),
            initial_price=Decimal(120),
            initial_shares=500,
        )
        a3_2 = asset_position_factory.create(
            underlying_instrument=e3,
            portfolio=portfolio,
            date=weekday,
            weighting=Decimal(0.1322),
            initial_price=Decimal(80),
            initial_shares=200,
        )

        price = instrument_price_factory.create(instrument=product, date=previous_sync_date, net_value=100)
        price_computation.compute(product, weekday, override_execution_datetime_validity=True)

        total_perf = (
            (a1_2._price / a1_1._price - 1) * a1_1.weighting
            + (a2_2._price / a2_1._price - 1) * a2_1.weighting
            + (a3_2._price / a3_1._price - 1) * a3_1.weighting
        )
        assert product.prices.count() == 2
        assert float(price.net_value * (Decimal(1.0) + total_perf)) == pytest.approx(
            float(product.prices.filter(date=weekday).first().net_value)
        )



import logging
import math
from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Any, Optional

import pandas as pd
from celery import shared_task
from celery.canvas import Signature
from django.db import models
from django.utils import timezone
from pandas.tseries.offsets import BDay
from wbfdm.models import Instrument
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

from .synchronization import SynchronizationTask


class PriceComputation(SynchronizationTask):
    @property
    def instruments(self):
        for instrument in Instrument.active_objects.filter(
            models.Q(id__in=self.products.values("id")) | models.Q(id__in=self.indexes.values("id"))
        ):
            yield instrument

    def compute(
        self,
        instrument: models.Model,
        sync_date: date,
        task_execution_datetime: Optional[datetime] = None,
        override_execution_datetime_validity: Optional[bool] = False,
        **kwargs: Any,
    ):
        """
        Given positions at t and t-1, we compute the performance and estimate the instrument price at t
        If a price already exists at that date and is estimated already, we update it. If no price exists, we create it as estimated.
        Otherwise, we don't do anything to protect imported real prices.
        :param instrument: The instrument to compute the new price from
        :param sync_date: The date at which we need to compute the new price
        :param task_execution_datetime: An optional datetime specifying at which time this task was initially executed.
        :param override_execution_datetime_validity: If true, we don't valide `task_execution_datetime`
        :param kwargs: keyword arguments
        """
        if not task_execution_datetime:
            task_execution_datetime = timezone.now()

        if (
            (self.is_valid_date(task_execution_datetime) or override_execution_datetime_validity)
            and instrument.is_active_at_date(sync_date)
            and not sync_date.weekday() in [5, 6]
        ):
            price_data = self._import_method(instrument, sync_date, **kwargs)
            if (
                price_data
                and (_instrument := price_data.pop("instrument", None))
                and (_date := price_data.pop("date", None))
            ):
                calculated = price_data.pop("calculated", True)
                try:
                    price = InstrumentPrice.objects.get(instrument=_instrument, date=_date, calculated=calculated)
                    for k, v in price_data.items():
                        setattr(price, k, v)
                    price.save()
                except InstrumentPrice.DoesNotExist:
                    price = InstrumentPrice.objects.create(
                        instrument=_instrument, date=_date, calculated=calculated, **price_data
                    )

                price.save()  # trigger explicitly save logic as update_or_create doesn't
                if (
                    _date == _instrument.prices.latest("date").date
                ):  # if price date is the latest instrument price date, we recomputet the last valuation data
                    _instrument.update_last_valuation_date()
        else:
            logging.info(
                f"Price Computation invalid: {str(instrument)} price computation with {self.name} was triggered for {sync_date} but date not valid for crontab schedule {str(self.crontab)}"
            )

    def compute_price_as_task_si(self, instrument: models.Model, sync_date: date, **kwargs: Any) -> Signature:
        """
        Utility function that returns the signature of the compute method
        """
        return compute_price_as_task.si(self.id, instrument.id, sync_date, **kwargs)

    def _tasks_signature(
        self, sync_date: Optional[date] = None, to_date: Optional[date] = None, **kwargs: Any
    ) -> Signature:
        """
        Gather all tasks that needs to run under this synchronization job as a list of celery signatures.
        This method is expected to be implemented at each inheriting class.
        :param args: list
        :param kwargs: dict
        :return: list[signature]
        """
        if not to_date:
            to_date = date.today()
        for instrument in self.instruments:
            # Get latest valuation date + 1 Bday if not given by the key word arguments
            instrument_sync_dates = []
            if sync_date:
                instrument_sync_dates = [sync_date]
            elif not sync_date and instrument.prices.exists() and (last_price := instrument.prices.latest("date")):
                instrument_sync_dates = map(
                    lambda x: x.date(),
                    pd.date_range(
                        max(last_price.date, to_date - timedelta(days=7)),
                        to_date,
                        freq="B",
                        inclusive="left",
                    ),
                )
            for instrument_sync_date in instrument_sync_dates:
                if instrument.is_active_at_date(instrument_sync_date):
                    yield compute_price_as_task.si(self.id, instrument.id, instrument_sync_date, **kwargs)

    @classmethod
    def _default_callback(cls, instrument: Instrument, val_date: date, **kwargs: Any):
        """
        Default NAV computation function. We simply compute the performance given two positions for two dates and estimate the new price
        based on the overall performance.
        TODO: If exit/buy of positions, is this function still correct?
        :param instrument: The instrument to compute the new price from
        :param val_date: The date at which we need to compute the new price
        :param kwargs: keyword arguments
        """
        if portfolio := instrument.portfolio:
            # check if the asset portfolio of this instruments exists and has positions at the synchronization date
            if previous_date := portfolio.get_latest_asset_position_date(val_date - BDay(1), with_estimated=True):
                # Get the previous valid price date before sync_date, checks if it exists and if positions are available
                # at that date.
                # If asset position exists on the previous day but not at the sync date, maybe propagation were not done, and we try it
                if portfolio.assets.filter(date=previous_date).exists():
                    portfolio.propagate_or_update_assets(previous_date, val_date)
                    if portfolio.assets.filter(date=val_date).exists():
                        last_price = None
                        if (
                            last_valuation := instrument.prices.filter(date=previous_date)
                            .order_by("calculated")
                            .first()
                        ):
                            last_price = last_valuation.net_value
                        elif not instrument.valuations.filter(date__lt=previous_date).exists():
                            last_price = instrument.issue_price
                        if last_price:
                            weights = pd.DataFrame(
                                portfolio.assets.filter(date=previous_date).values(
                                    "weighting", "date", "underlying_instrument"
                                )
                            )
                            weights = weights.pivot_table(
                                index="date", columns=["underlying_instrument"], values="weighting", aggfunc="sum"
                            ).astype("float")
                            weights = weights.iloc[-1, :]
                            perfs = pd.DataFrame(
                                portfolio.assets.filter(date__in=[previous_date, val_date]).values(
                                    "date", "price_fx_portfolio", "underlying_instrument"
                                )
                            )
                            perfs = perfs.pivot_table(
                                index="date",
                                columns=["underlying_instrument"],
                                values="price_fx_portfolio",
                                aggfunc="mean",
                            ).astype("float")
                            perfs = perfs / perfs.shift(1, axis=0) - 1.0
                            perfs = perfs.fillna(0).iloc[-1, :]
                            total_perfs = float((perfs * weights).sum())
                            new_gross_valuation = float(last_price) * (1.0 + total_perfs)
                            if new_gross_valuation and not math.isnan(new_gross_valuation):
                                return {
                                    "instrument": instrument,
                                    "date": val_date,
                                    "gross_value": Decimal(new_gross_valuation),
                                    "net_value": Decimal(new_gross_valuation),
                                }

        return None

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:pricecomputationrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


@shared_task(queue="portfolio")
def compute_price_as_task(price_computation_method_id: int, instrument_id: int, sync_date: date, **kwargs: Any):
    instrument = Instrument.objects.get(id=instrument_id)
    price_computation = PriceComputation.objects.get(id=price_computation_method_id)
    price_computation.compute(instrument, sync_date, **kwargs)
